#!/usr/local/bin/perl

#use Strict;
#use Warnings;

# perl antispam smtp proxy
# (c) John Hanna, John Calvi, Robert Orso, AJ 2004 under the terms of the GPL
# ASSP founded and developed to Version 1.0.12 by John Hanna.
# ASSP web interface by AJ.
# ASSP development since 1.0.12 by John Calvi.
# LDAP implementation by Robert Orso.

# Special Thanks for contributions to.....
  # Nigel Barling - SPF & RBL.
  # Mark Pizzolato - SMTP Session Limits.

use bytes; # get rid of annoying 'Malformed UTF-8' messages

$version='1.1.1';
$modversion='Final'; #appended in version display.

$CanUseLDAP = eval("use Net::LDAP; 1"); # Net LDAP module installed
$CanUseAddress = eval("use Email::Valid; 1"); # Email Valid module installed
$CanUseDNS = eval("use Net::DNS; 1"); # Net DNS module installed - required for SPF & RBL
$AvailSPF = eval("use Mail::SPF::Query; 1"); # Mail SPF module installed
$CanUseSPF = $AvailSPF && $CanUseDNS; # SPF and dependancies installed
$CanUseRBL = $CanUseDNS; # RBL and dependancies installed

# load from command line if specified
if($ARGV[0]) {
 $base=$ARGV[0];
} else {
 # the last one is the one used if all else fails
 for ('.','assp','/usr/local/assp','/home/assp','/usr/assp','/assp','.') {
  $base=$_;
  last if -e "$base/assp.cfg";
 }
}

loadConfig();
sub loadConfig {
 print "loading config -- base='$base'\n";
 @Config=(
 [0,0,0,heading,'Network Setup'],
 # except for the heading lines, all config lines have the following:
 #  $name,$nicename,$size,$func,$default,$valid,$onchange,$description
 # name is the variable name that holds the data
 # nicename is a human readable pretty display name (oh how nice!)
 # size is the appropriate input box size
 # func is a function called to render the config item
 # default is the default value
 # valid is a regular expression used to clean and validate the input -- no match is an error and $1 is the desired result
 # onchange is a function to be called when this value is changed -- usually undef; just updating the value is enough
 # group is the heading group belonged to.
 # description is text displayed to help the user figure what to put in the entry

 [smtpDestination,'SMTP Destination',20,textinput,'127.0.0.1:225','(\S*)',undef,
  'The address:port of your message handling system\'s smtp server. For example: 127.0.0.1:125'],
 [AsAService,'As a Service',0,checkbox,'','(\S*)',undef,
  'In Windows 2000 / NT you can run it as a service; requires <a href="http://www.roth.net/perl/Daemon/" rel="external">win32::daemon</a>. Requires start from the service control panel.'],
 [AsADaemon,'As a Daemon',0,checkbox,1,'(\S*)',undef,
 'In Linux/BSD/Unix/OSX fork and close file handles, kinda like "perl assp.pl &amp;" but better. Requires restart.'],
 [myName,'My Name',20,textinput,'ASSP-nospam','(\S+)',undef,
  'What the program calls itself in the email "received by" header. Usually ASSP-nospam.'],
 [listenPort,'Listen Port',20,textinput,'125','(\S+)',ConfigChangeMailPort,
  'On what port should ASSP accept smtp connections? Normally 25. You can supply an interface:port to limit connections.'],
 [listenPort2,'Another Listen Port',20,textinput,'','(\S+)',ConfigChangeMailPort2,
  'Listen for incoming SMTP requests on a second port.<br />
  For those who cannot use SMTP Port 25 outside of their ISP Network,
  or as a dedicated port for VPN purposes.<br />
  You can supply an interface:port to limit connections. For example: 2525 or 127.0.0.2:325'],
 [maxSMTPSessions,'Maximum SMTP Sessions',20,textinput,'','(\S+)',undef,
  'The maximum number of SMTP sessions (connections) to handle concurrently.<br /> 
  This can help if the server is overloading. 20 simultaneous connections is typically enough.
  No entry = no limit.'],
 [maxSMTPipSessions,'Maximum Sessions/IP',20,textinput,'','(\S+)',undef,
  'The maximum number of SMTP sessions (connections) to handle per IP address concurrently.<br />  
  Limit this to prevent DOS attacks, 5 simultaneous connections is typically enough.
  No entry = no limit.<br /> 
  Also note that the ISP &amp; accept all mail addresses are excluded from limiting.'],
 [SessionLog,'Session Limit Logging',0,checkbox,'','(\S*)',undef,
  'Log an event each time the above session limits are triggered.'],
 [proxyserver,'Proxy Server',20,textinput,'','(\S*)',undef,
  'Use a Proxy Server for up/downloading greylist etc. Format - interface:port. 
  For example: 192.168.0.1:8080'],
 [webAdminPort,'Web Admin Port',20,textinput,55555,'(\S+)',ConfigChangeAdminPort,
  'On what port should ASSP listen for http connections for the web administration interface?<br />
  Changing this will require changing the URL on your browser to reconnect.
  You can supply an interface:port to limit connections.'],

[0,0,0,heading,'SPAM Control'],
 [SpamError,'Spam Error',80,textinput,'500 Mail appears to be unsolicited -- send error reports to postmaster@yourdomain.com','(5\d\d .*)',undef,
  'SMTP error message to reject spam.'],
 [spamaddresses,'Spam Addresses*',60,textinput,'put|your@spambucket.com|addresses|@here.org','(.*)',ConfigMakeRe,
  'Mail to any of these users at are always spam unless from someone on the whitelist;<br />
  @domain.com makes the whole domain a spam domain. A username without domain will register across all local domains.'],
 [spamLovers,'Spam-Lover Addresses*',60,textinput,'postmaster','(.*)',ConfigMakeRe,
  'Spam addressed entirely to spam lovers is not blocked. Mail addressed to both spam lovers and non spam lovers IS blocked.<br />
  Accepts specific addresses (user@domain.com), addresses at local domains (user), or entire local domains (@domain.com).<br />
  Separate entries with pipes: |. For example: jhanna@thisdomain.com|fhanna|@sillyguys.org'],
 [noProcessing,'Unprocessed Addresses*',40,textinput,'','(.*)',ConfigMakeRe,
  'Mail solely to or from any of these addresses are ignored by ASSP.<br />
  Like a more efficient version of spamLovers &amp; redlist combined. Valid entry types are as per spamlovers.'],
 [blackListedDomains,'Blacklisted Domains*',60,textinput,'','(.*)',ConfigMakeRe,
  'Domains from which you always want to reject mail, they only send you spam. For example: spam.net|xxxpics.com'],
 [whiteRe,'Expression to Identify Non-Spam',60,textinput,'','(.*)',ConfigCompileRe,
  'If an incoming email matches this Perl regular expression it will be considered non-spam.<br />
  For example: Secret Ham Password|307\D{0,3}730\D{0,3}4[12]\d\d<br />
  For help writing regular expressions click <a href="http://www.perlmonks.org/index.pl?node=perlre" rel="external">here</a>.
  Note that flags are "si" and the header as well as body is scanned.<br /> 
  Some things you might include here are your office phone number or street address, spam rarely includes these details.'],
 [blackRe,'Expression to Identify Spam',60,textinput,'http://[\\w\\.]+@|\w<[a-z0-9]+[abcdfghjklmnpqrstuvwxyz0-9]{4}[a-z0-9]*>|subject: [^\\n]*     \S','(.*)',ConfigCompileRe,
  'If an incoming email that\'s not local or whitelisted matches this Perl regular expression it will be considered spam.<br />
  May match text from the body or header of the email. For example: penis|virgin|X-Priority: 1'],
 [redRe,'Expression to Identify Redlisted Mail',60,textinput,'','(.*)',ConfigCompileRe,
  'If an email header matches this Perl regular expression it will be considered redlisted. For example: \\[autoreply\\]'],
 [npRe,'Expression to Identify No-processing Mail',60,textinput,'','(.*)',ConfigCompileRe,
  'If an email header matches this Perl regular expression it will pass through unprocessed. For example: \\[autoreply\\]'],
 [heloBlacklistIgnore,'Don\'t block these HELO\'s*',60,textinput,'','(.*)',ConfigMakeRe,
  'HELO / EHLO greetings on this list will be excluded from the HELO blacklist. For example: host123.isp.com|host456.isp.com'],
 [noHeloBlacklist,'Don\'t Use the Helo Blacklist',1,checkbox,'','(.*)',undef,
  'Check this box to accept mail from blacklisted-helo hosts.<br />
  You probably want to disable the HELO blacklist in the initial training phase for ASSP.'],
 [noGreyListUpload,'Don\'t Upload Greylist Stats',0,checkbox,'','(.*)',undef,
  'Check this to disable the greylist upload when rebuildspamdb runs.'],
 [nogreydownload,'Don\'t auto-download the greylist file',0,checkbox,'','(.*)',undef,'Set this checkbox if don\'t use the greylist or want to download it manually.'],
 [AddSpamProbHeader,'Add Spam Probability Header?',0,checkbox,1,'(.*)',undef,
  'Adds a line to the email header "X-Assp-Spam-Prob: 0.0123" Probs range from 0 to +1 where > 0.6 = spam.'],
 [NoExternalSpamProb,'Block Outgoing Spam-Prob header?',0,checkbox,1,'(.*)',undef,
  'Check this box if you don\'t want your X-Assp-Spam-Prob header on external mail<br />
  Note this means mail from local users to local users will also be missing the header.'],
 [AddSpamHeader,'Add Spam Header?',0,checkbox,1,'(.*)',undef,
  'Adds a line to the email header "X-Assp-Spam: YES" if the message is spam.'],
 [UseSubjectsAsMaillogNames,'Use Subject as Maillog Names',0,checkbox,1,'(.*)',undef,
  'You can turn this on to help you manually identify mail in your spam and non-spam collections.'],

[0,0,0,heading,'Whitelist Options'],
 [whiteListedDomains,'Whitelisted Domains*',60,textinput,'sourceforge.net','(.*)',ConfigMakeRe,
  'Domains from which you want to receive all mail<br />
  Your ISP, domain registration, mail list servers, stock broker, or other key business partners might be good candidates.<br />
  Note this matches the end of the address, so if you don\'t want to match subdomains then include the @.<br />
  Note that buy.com would also match spambuy.com but .buy.com won\'t match buy.com.<br />
  DO NOT put your local domains on this list. For example: sourceforge.net|@google.com|.buy.com'],
 [MaxWhitelistDays,'Max Whitelist Days',5,textinput,'90','(\d+)',undef,
  'This is the number of days an address will be kept on the whitelist without any email to/from this address.'],
 [WhitelistOnly,'Reject All But Whitelisted Mail',0,checkbox,'','(.*)',undef,
  'Check this if you don\'t want Bayesian filtering and want to reject all mail from anyone not whitelisted.<br />
  Note: this turns the redlist into a blacklist.'],
 [NoMaillog,'Don\'t log mail',0,checkbox,'','(.*)',undef,
  'Check this if you\'re using Whitelist-Only and don\'t care to save mail to build the Bayesian database.'],
 [NotGreedyWhitelist,'Only the envelope-sender is added/compared to the whitelist',0,checkbox,'','(.*)',undef,
  'Normal operation includes addresses in the FROM, SENDER, REPLY-TO, ERRORS-TO, or LIST-* header fields.<br />
  This allows nearly all list email to be whitelisted. Check this option to disable this.'],
 [WhitelistLocalOnly,'Only local or authenticated users contribute to the whitelist.',0,checkbox,'','(.*)',undef,
  'Normal operation allows all local, authenticated, or whitelisted users to add to the whitelist.<br />
  Check this box to not allow whitelisted users to add to the whitelist.'],
 [KeepWhitelistedSpam,'Keep Whitelisted Spam',0,checkbox,0,'(.*)',undef,
  'Check this box if you don\'t want rebuildspamdb to remove entries from the spam collection
   after subsequent whitelisting.<br /> Checking this box will speed up your rebuild.'],

 [0,0,0,heading,'Relaying'],
 [ispip,'ISP/Secondary MX Servers',40,textinput,'127.0.0.|10.|169.254.|172.16.|192.168','(\S*)',ConfigMakeRe,
  'Enter any addresses that are your ISP or backup MX servers, separated by pipes (|).<br />
  These addresses will (necessarily) bypass Greylist, ip Limiting,  SPF &amp; RBL checks.<br />
  Do not leave this blank -- enter "none" for none. For example: 127.0.0.1|10.'],
 [ispgreyvalue,'ISP/Secondary MX Grey Value',10,textinput,'','(\S*)',undef,
  'It is recommended that for ISP &amp; Secondary MX servers to bypass their greylist values<br />
  For eg. 0.5 (Completely grey). If left blank then the greylist "X" value is used.<br />
  Note: value should be greater than 0 and less than 1, where 0 = never spam &amp; 1 = always spam'],
 [acceptAllMail,'Accept All Mail*',40,textinput,'127.0.0.|10.|169.254.|172.16.|192.168.','(.*)',ConfigMakeRe,
  'Denies relaying for hosts besides these. These hosts also contribute to the whitelist.<br />
  Do not leave this blank -- enter "none" for none. For example: 127.0.0.1|10.|169.254.|172.16.|192.168.'],
 [relayHostFile,'Relay Host File',40,textinput,'','(.*)',undef,
  'Like Accept All Mail, but this is a file that contains a list of ip addresses (one per line)
   for whom you want to relay mail.<br />
   This is an ABSOLUTE path, not relative to base. For example: /usr/local/assp/relayhosts'],
 [localDomains,'Local Domains*',80,textinput,'putYourDomains.com|here.org','(.*)',ConfigMakeRe,
  'Addresses in these domains are considered local delivery. Separate addresses with |. Include all subdomains.<br />
  Don\'t leave this blank -- if you don\'t have any local domains then use xyzxyz. For example: put.YourDomains.com|here.org'],
 [localDomainsFile,'Local Domains File',40,textinput,'','(.*)',undef,
  'Like Local Domains, but this is a file that contains a list of host names (one per line) for whom you want to accept mail.<br />
  This is an ABSOLUTE path, not relative to base. For example: /usr/local/assp/locals'],
 [PopB4SMTPFile,'Pop Before SMTP DB File',40,textinput,'','(.*)',undef,
  'Enter the DB database filename of your POP before SMTP implementation with records stored for 
  dotted-quad IP addresses<br />
  For example: /etc/mail/popip.db If it\'s got something else, you\'ll need to edit the PopB4SMTP subroutine.'],
 [relayHost,'Relay Host',40,textinput,'','(\S*)',undef,
  'Your isp\'s mail relayhost (smarthost). For example: mail.isp.com:25<br />
  If you run Exchange/Notes and you want assp to update the nonspam database and the whitelist,<br />
  then enter your isp\'s smtp relay host here. Blank means no relayhost. Only required if clients don\'t deliver through SMTP.'],
 [relayPort,'Relay Port',20,textinput,'','(\S*)',ConfigChangeRelayPort,
  'Tell your mail server to connect to this port (e.g. 127.0.0.1:225 or 127.0.0.2) as its smarthost / relayhost.
  For example: 225<br /> Note that you\'ll want to keep the relayPort protected from external access by your firewall.<br />
  You can supply an interface:port to limit connections.'],
 [NoRelaying,'No Relaying Error',80,textinput,'550 Relaying not allowed','(5\d\d .*)',undef,
  'SMTP error message to deny relaying.'],
 [defaultLocalHost,'Default Local Host',40,textinput,'','(\S*)',undef,
  'If you want to be able to send mail to local users without a domain name then put the 
  default local domain here.<br /> Blank disables this feature. For example: mydomain.com'],

[0,0,0,heading,'Validate Recipients'],
 [NoValidRecipient,'No-Valid-User Reply',80,textinput,'550 5.1.1 User unknown','([5|2]\d\d .*)',undef,
  'SMTP reply for invalid Users. Default: \'550 5.1.1 User unknown\'<br />
  You may reply with a \'fake OK\' by entering \'250 OK - Recipient &lt;EMAILADDRESS&gt;\' 
  to confuse address harvesters.<br />
  The literal EMAILADDRESS (case sensitive) is replaced by the fully qualified SMTP recipient (e.g. thisuser@yourcompany.com).'],
 [DoRFC822,'Validate recipient addresses to conform with RFC 822',0,checkbox,1,'(.*)',undef,
  'If activated, each recipient address is checked to conform with the email format defined in RFC 822.<br />
  This requires an installed <a href="http://search.cpan.org/~maurice/Email-Valid-0.15/Valid.pm" rel="external">Email::Valid</a> module in PERL.'],
 [LocalAddresses_Flat,'Local Addresses*',40,textinput,'','(.*)',ConfigMakeRe,
  'These email addresses are considered local by ASSP. You can list specific addresses (user@mydomain.com),<br />
  addresses at any local domain (user), or entire local domains (@mydomain.com). Separate entries with pipes: |.<br />
  For example: jhanna@thisdomain.com|fhanna|@sillyguys.org or place them in a plain ASCII file one address per line.'],
 [DoLDAP,'Do LDAP lookup for valid recipients',0,checkbox,1,'(.*)',undef,
  'Check recipients against an LDAP database before accepting the message.<br />
  Note: Checking this requires filling in all the other values below.
  This requires an installed <a href="http://search.cpan.org/~gbarr/perl-ldap-0.31/lib/Net/LDAP.pod" rel="external">NET::LDAP</a> module in PERL.'],
 [LDAPHost,'LDAP Host',60,textinput,'localhost','(\S*)',undef,
  'Enter the DNS-name or IP address of the server that runs the LDAP database. For example: localhost'],
 [LDAPLogin,'LDAP Login',60,textinput,'','(.*)',undef,
  'Most LDAP servers require a login and password before they allow queries.<br /> 
  Enter the DN specification for a user with sufficient permissions here.<br />
  For example: cn=Administrator,cn=Users,DC=yourcompany,DC=com'],
 [LDAPPassword,'LDAP Password',2,textinput,'','(.*)',undef,
  'Enter the password for the specified LDAP login here.'],
 [LDAPRoot,'LDAP Root container',60,textinput,'','(.*)',undef,
  'The LDAP lookup will use this container and all sub-containers to match the query.<br />
  For example: DC=yourcompany,DC=com'],
 [LDAPFilter,'LDAP Filter',60,textinput,'','(\S*)',undef,
  'This filter is used to query the LDAP database. This strongly depends on the LDAP structure.<br />
  The filter must return an entry if the recipient address matches with that of any user.<br />
  The literal EMAILADDRESS is replaced by the fully qualified SMTP recipient (eg. user@domain.com) during the search.<br /> 
  For example: (proxyaddresses=smtp:EMAILADDRESS)'],

  [0,0,0,heading,'SPF Options'],
 [ValidateSPF,'Enable SPF Validation',0,checkbox,1,'(.*)',undef,
  'Enable Sender Policy Framework Validation as described at <a href="http://spf.pobox.com">spf.pobox.com</a>.<br />
  This requires an installed <a href="http://spf.pobox.com/downloads.html" rel="external">Mail::SPF::Query</a> module in PERL.<br />
  Senders that fail SPF validation will receive a 550 5.7.1 [Sender not authorized by domain policy] SMTP error code.<br />
  Note: no error is sent if SPF is in test mode.'],
 [SPFWL,'Whitelisted SPF Validation',0,checkbox,1,'(.*)',undef,
  'Enable Sender Policy Framework Validation for whitelisted users also.'], 
 [LocalPolicySPF,'Local SPF Policy',40,textinput,'v=spf1 a/24 mx/24 ptr ~all','(.*)',undef,
  'If the sending domain does not publish its own SPF Records a local policy can be defined.<br />
  The default is v=spf1 a/24 mx/24 ptr ~all'],  
 [AddSPFHeader,'Add Received-SPF Header',0,checkbox,1,'(.*)',undef,
  'Add Received-SPF header to header of all emails processed by SPF.'], 
 [SPFLog,'Enable SPF logging',0,checkbox,0,'(.*)',undef,
  'Enables verbose logging of all SPF checks in the maillog. Default is to log failures only.'],
 [DebugSPF,'Enable SPF Debug output to ASSP Logfile',0,checkbox,0,'(.*)',undef,
  'Enables verbose debugging of SPF queries within the Mail::SPF::Query module.'],

  [0,0,0,heading,'RBL Options'],
 [ValidateRBL,'Enable Realtime Blacklist Validation',0,checkbox,1,'(.*)',updateRBL,
  'Senders that fail RBL validation will receive a 550 5.7.1 [Sender not authorized by domain policy] SMTP error code.<br />
  Note: no error is sent if RBL is in test mode.'],  
 [RBLWL,'Whitelisted RBL Validation',0,checkbox,1,'(.*)',undef,
  'Enable RBL for whitelisted users also'], 
 [AddRBLHeader,'Add X-Assp-Received-RBL Header',0,checkbox,1,'(.*)',undef,
  'Add X-Assp-Received-RBL header to header of all emails processed by RBL.'],
 [RBLLog,'Enable RBL logging',0,checkbox,0,'(.*)',undef,
  'Enables verbose logging of all RBL checks in the maillog. Default is to log failures only.'],
 [RBLServiceProvider,'RBL Service Providers',80,textinput,'bl.spamcop.net|cbl.abuseat.org|sbl-xbl.spamhaus.org|dnsbl.njabl.org|list.dsbl.org|dnsbl.sorbs.net|opm.blitzed.org|dynablock.njabl.org','(\S*)',updateRBLSP,
  'Domain Names of RBLs to use separated by "|". Defaults are...<br />
  bl.spamcop.net|cbl.abuseat.org|sbl-xbl.spamhaus.org|dnsbl.njabl.org|list.dsbl.org|dnsbl.sorbs.net|opm.blitzed.org|dynablock.njabl.org'],   
 [RBLmaxreplies,'Maximum Replies',1,textinput,6,'(\d*)',updateRBLMR,'A reply is affirmative or negative reply from a RBL.<br />
  The RBL module will wait for this number of replies (negative or positive) from the RBLs listed under Service Provider for up to the Maximum Time below.<br />
  This number should be equal to or less than the number of RBL Service Providers listed to allow for randomly unavailable RBLs'],
 [RBLmaxhits,'Maximum Hits',1,textinput,3,'(\d*)',updateRBLMH,'A hit is an affirmative response from a RBL.<br />
  The RBL module will check all of the RBLs listed under Service Provider,<br />
  and flag the email with a RBL failure flag if more than this number of RBLS return a postive blacklisted response.<br />
  This number should be less than or equal to Maximum Replies above and greater than 0'],
 [RBLmaxtime,'Maximum Time',1,textinput,10,'(\d*)',undef,'This sets the maximimum time to spend on each message performing RBL checks'],
 [RBLsocktime,'Socket Timeout',1,textinput,5,'(\d*)',undef,'This sets the RBL socket read timeout'],

[0,0,0,heading,'Virus Control'],
 [BlockExes,'External Attachment Blocking',1,textinput,0,'(\d*)',undef,
  'Enter a number from 0-3 to determine which level of attachment protection to provide for external mail.<br />
  Enter 0 or leave blank for no attachment blocking.'],
 [BlockWLExes,'Whitelisted &amp; Local Attachment Blocking',1,textinput,0,'(\d*)',undef,
  'Enter a number from 0-3 to determine which level of attachment protection for whitelisted &amp; local senders.'],
 [BlockNPExes,'NoProcessing Attachment Blocking',1,textinput,0,'(\d*)',undef,
  'Enter a number from 0-3 to determine which level of attachment protection for no processing senders.'],
 [BadAttachL1,'Level 1 Blocked File Extensions',60,textinput,'','(.*)',updateBadAttachL1,
  'This regular expression is used to identify Level 1 attachments that should be blocked.<br />
  Separate entries with a pipe |. The dot . is assumed to preceed these, so don\'t include it. For example:<br />
  ad[ep]|asx|ba[st]|chm|cmd|com|cpl|crt|dbx|exe|hlp|ht[ab]|in[fs]|isp|js|jse|lnk|md[abez]|mht|ms[cipt]|nch|pcd|pif|prf|reg|sc[frt]|sh[bs]|vb|vb[es]|wms|ws[cfh]'],
 [BadAttachL2,'Level 2 Blocked File Extensions',60,textinput,'','(.*)',updateBadAttachL2,
  'This regular expression is used to identify Level 2 attachments that should be blocked.<br />
  Level 2 already includes all blocked extensions from Level 1. For example:<br />
 (ad[ep]|asx|ba[st]|chm|cmd|com|cpl|crt|dbx|exe|hlp|ht[ab]|in[fs]|isp|js|jse|lnk|md[abez]|mht|ms[cipt]|nch|pcd|pif|prf|reg|sc[frt]|sh[bs]|vb|vb[es]|wms|ws[cfh]).zip'],
 [BadAttachL3,'Level 3 Blocked File Extensions',60,textinput,'zip','(.*)',updateBadAttachL3,
  'This regular expression is used to identify Level 3 attachments that should be blocked.<br />
  Level 3 includes Level 2 and Level 1. For example: zip|url'],
 [AttachmentError,'Attachment Error',80,textinput,'500 These attachments are not allowed -- Compress before mailing.','(5\d\d .*)',undef,
  'SMTP error message to reject attachments.'],
 [AvPath,'Path to Anti-virus Databases',60,textinput,'','(.*)',undef,
  'The directory path to your anti-virus databases, uses ASSP\'s base if left blank. For example: /usr/share/clamav/db'],
 [AvDbs,'List of Anti-virus Signature Database Files',60,textinput,'main.db,daily.db','(.*)',undef,
  'A comma (no space!) separated list of virus signature files. Blank this out to disable virus scanning.<br />
  For example: main.db,daily.db. These files are available on http://assp.sourceforge.net<br />
  NOTE: Changing this field requires restarting ASSP before changes take effect.'],
 [AvError,'Error Message to Reject Infected Email',60,textinput,'500 Mail appears infected with \'$infection\' -- disinfect and resend.','(5\d\d .*)',undef,
  'SMTP error message to reject infected mail. The string $infection is replaced with the name of the detected virus.<br />
  For example: 500 Mail appears infected with \'$infection\' -- disinfect and resend.'],
 [Avlocal,'Virus Scan Local',0,checkbox,1,'(.*)',undef,
  'Check this box to scan local users email as well.'],
 [AVBytes,'AV Bytes',10,textinput,100000,'(\d*)',undef,
  'How many bytes of the message will be Virus scanned? For example: 100000<br />
  Leave Blank to scan entire email which results in a significant performance penalty on large attachments.<br />
  Most virus signatures match in the first 20-100K of the message.'],

[0,0,0,heading,'Spam Bombs &amp; Scripting'],
 [bombRe,'Expression to Identify Spam Bombs',60,textinput,'','(.*)',ConfigCompileRe,
  'It is possible for a spammer to create 1000s of messages that appear to be from your domain.<br />
  When these messages bounce, the bounces come to you. You can use this feature to block those messages.<br />
  Leave this blank to disable the feature. For example: images/ad12\.gif'],
 [bombError,'Spam Bomb Error',80,textinput,'500 Your message was rejected because it appears to be part of a spam bomb -- rephrase your message and try sending it again.','(.*)',undef,
  'SMTP error message to reject spam bombs.'],
 [scriptRe,'Expression to Identify Mobile Scripts',60,textinput,'','(.*)',ConfigCompileRe,
  'Spam emails may contain mobile scripting code, eg activex and java. You can use this feature to block those messages.<br />
  Leave this blank to disable the feature. For example: \&lt;applet|\&lt;embed|\&lt;iframe|\&lt;object|\&lt;script|onmouseover|javascript:'],
 [scriptError,'Script Error',80,textinput,'500 Your email contains html scripting code -- please resend as plain text.','(.*)',undef,
  'SMTP error message to reject scripts.'],

[0,0,0,heading,'TestMode Options'],
 [baysTestMode,'Bayesian Test Mode',0,checkbox,1,'(.*)',undef,
  'If set then all Bayesian Spam messages are delivered.'],
 [blTestMode,'Blacklist Test Mode',0,checkbox,1,'(.*)',undef,
  'If set then all Blacklisted messages are delivered.'],
 [hlTestMode,'Helolist Test Mode',0,checkbox,1,'(.*)',undef,
  'If set then all Helolisted messages are delivered.'],
 [sbTestMode,'Spam Address Test Mode',0,checkbox,1,'(.*)',undef,
  'If set then all Spam Address messages are delivered.'],
 [spfTestMode,'SPF Test Mode',0,checkbox,1,'(.*)',undef,
  'If set then all SPF failed messages are delivered.'],
 [rblTestMode,'RBL Test Mode',0,checkbox,1,'(.*)',undef,
  'If set then all RBL failed messages are delivered.'],
 [spamSubject,'Prepend Spam Subject',10,textinput,'','(.*)',undef,
  'If TestMode and message is spam spamSubject gets prepended to the subject of the email. For example; [SPAM]'],
 [sendAllSpam,'Address to CC All Spam',20,textinput,'','(.*)',undef,
  'If you put an email address in this box ASSP will try to deliver a copy of all spam email to this address.<br />
  This is the forward all spam feature. For example: spammeister@mydomain.com'],

[0,0,0,heading,'Email Interface'],
 [EmailInterfaceOk,'Enable Email Interface',0,checkbox,1,'(.*)',undef,
  'Checked means that you want to ASSP to intercept and parse mail to the following usernames at any localdomains.<br />
  If you are using RelayHost and RelayPort see 
  <a href="http://assp.sourceforge.net/fom/cache/45.html" rel="external">this note</a>.'],
 [EmailSpam,'Report Spam Address',20,textinput,'assp-spam','(.*)',undef,
  'Any mail sent by local/authenticated users to this username will be interpreted as a spam report.<br />
  No mail is delivered! For example: assp-spam'],
 [EmailHam,'Report not-Spam Address',20,textinput,'assp-notspam','(.*)',undef,
  'Any mail sent by local/authenticated users to this username will be interpreted as a false-positive report.<br />
  No mail is delivered! For example: assp-notspam'],
 [EmailWhitelist,'Add to Whitelist Address',20,textinput,'assp-white','(.*)',undef,
  'Any mail sent by local/authenticated users to this username will be interpreted as a request to add addresses to the whitelist.<br /> No mail is delivered! For example: assp-white'],
 [EmailFrom,'From Address for Email',20,textinput,'ASSP <postmaster@yourdomain.com>','(.+)',undef,
  'Email sent from ASSP acknowledging your submissions will be sent from this address.<br />
  Some mailers don\'t like the default setting. For example: ASSP &lt;&gt; or Mail Administrator
  &lt;mailadmin@mydomain.com&gt;'],
 [NoHaiku,'Don\'t reply to messages to the Email Interface',0,checkbox,0,'(.*)',undef,
  'Check this option to suppress email reports for spam and not-spam reports and whitelist additions
  via the email interface.'],

[0,0,0,heading,'File Paths'],
 ['base','Directory Base',40,textinput,'.','',undef,'All paths are relative to this folder.<br /> 
 <b>Note: this must be changed as a command line parameter and is displayed here for reference only.</b>'],
 [spamlog,'Spam Collection',40,textinput,'spam','(\S+)',undef,'The folder to save the collection of spam emails. For example: spam'],
 [notspamlog,'Not-spam Collection',40,textinput,'notspam','(\S+)',undef,'The folder to save the collection of not-spam emails. For example: notspam'],
 [incomingOkMail,'External OK mail',40,textinput,'','(.*)',undef,'The folder to save Bayesian non-spam (message ok).
 Leave this blank to not save these files (default).<br />
 If you want to keep copies of OK mail then put in a directory name. Note: you must create the directory. For example: okmail'],
 [viruslog,'Virus Collection',40,textinput,'','(.*)',undef,
'The folder to save virii, blocked attachments and scripting. Leave this blank to not save these files (default).<br />
 If you want to keep copies of blocked content then put in a directory name. Note: you must create the directory. For example: virii'],
 [correctedspam,'False-negative Collection',40,textinput,'errors/spam','(\S+)',undef,
'Spam that got through -- counts double. For example: errors/spam'],
 [correctednotspam,'False-positive Collection',40,textinput,'errors/notspam','(\S+)',undef,
'Good mail that was listed as spam, count 4x. For example: errors/notspam'],
 [maillogExt,'Extension for Mail Files',10,textinput,'.eml','(\S*)',undef,
'Enter the file extension (include the period) you want appended to the mail files in the mail collections.<br />
 Leave it blank for no extension. For Example: .eml'],
 [spamdb,'Spam Bayesian Database File',40,textinput,'spamdb','(\S+)',undef,'The output file from rebuildspamdb.pl.'],
 [whitelistdb,'Email Whitelist Database File',40,textinput,'whitelist','(\S+)',undef,'The file with the whitelist.'],
 [redlistdb,'Email Redlist Database File',40,textinput,'redlist','(\S+)',undef,'The file with the redlist.'],
 [dnsbl,'DNS Blacklist Database File',40,textinput,'','(\S*)',undef,'The file with the current DNSBL -- make this blank if you don\'t use it.'],
 [greylist,'Greylist Database',40,textinput,'greylist','(\S*)',undef,'The file with the current greylist database -- make this blank if you don\'t use it.'],
 [logfile,'ASSP Logfile',40,textinput,'maillog.txt','(\S*)',ConfigChangeLogfile,
'Blank if you don\'t want a log file. Change it to maillog.log if you don\'t want auto rollover.'],
 [pidfile,'PID File',40,textinput,'pid','(\S*)',undef,'Blank to skip writing a pid file. *nix users need pid files.
 Leave it blank in Windows.<br /> You have to restart the service before you get a pid file in the new location.'],

[0,0,0,heading,'Logging'],
 [silent,'Silent Mode',0,checkbox,'','(.*)',undef,
  'Checked means don\'t print log messages to the console. AsADaemon overrides this.'],
 [DEBUG,'Debug Mode',0,checkbox,'','(.*)',ConfigDEBUG,
  'Checked sends debugging info to a .dbg file. 
 Leave this unchecked unless there is a program error you are trying to track down.'],
 [sendNoopInfo,'Send NOOP Info',0,checkbox,'','(.*)',undef,
  'Checked means you want ASSP to send a "NOOP Connection from $ip" message to your SMTP server. (Postfix croaks on this.)'],
 [wlAttachLog,'Whitelisted Blocked Attachments',1,textinput,5,'(\d*)',undef,
'Where to store Whitelisted blocked attachments. Recommended : 5 <br /><span class="negative"> 2 = notspam folder, 3 = spamfolder, 4 = mailok folder, 5 = virii folder, 6 = discard, 7 = discard (&amp; cc to spam account).</span><br />'],
 [npAttachLog,'No Processing Blocked Attachments',1,textinput,5,'(\d*)',undef,'Where to store external blocked attachments. Recommended: 5'],
 [extAttachLog,'External Blocked Attachments',1,textinput,5,'(\d*)',undef,'Where to store external blocked attachments. Recommended: 5'],
 [spamBombLog,'Spam Bombs',1,textinput,6,'(\d*)',undef,'Where to store spam bombs. Recommended: 6'],
 [scriptLog,'Scripts',1,textinput,3,'(\d*)',undef,'Where to store scripted emails. Recommended: 3'],
 [baysNonSpamLog,'Bayesian Non Spam',1,textinput,4,'(\d*)',undef,'Where to store bayesian non spam (message ok) emails. Recommended: 4'],
 [blDomainLog,'Blacklisted Domains',1,textinput,3,'(\d*)',undef,'Where to store blacklisted domain emails. Recommended: 3'],
 [spamHeloLog,'Spam Helos',1,textinput,3,'(\d*)',undef,'Where to store spam helo emails. Recommended: 3'],
 [spamBucketLog,'Spam Addresses',1,textinput,6,'(\d*)',undef,'Where to store has spam address emails. Recommended: 6'],
 [baysSpamLog,'Bayesian Spams',1,textinput,3,'(\d*)',undef,'Where to store bayesian spam emails. Recommended: 3'],
 [SPFFailLog,'SPF Failures',1,textinput,3,'(\d*)',undef,'Where to store SPF Failure spam emails. Recommended: 3'],
 [RBLFailLog,'RBL Failures',1,textinput,3,'(\d*)',undef,'Where to store RBL Failure spam emails. Recommended: 3'],
 [freqNonSpam,'Non Spam Log Frequency',1,textinput,1,'(\d*)',updateLog2,'Store every n\'th non spam message. 
eg. If you set the value to 10 then every 10th message is logged.<br />
These frequency settings are for ASSP users with a mature installtion who experience heavy mail or spam volumes.<br /> 
Enter a larger value if the non spam corpus is being refreshed too quickly. Default Value = 1, log every message.'],
 [freqSpam,'Spam Log Frequency',1,textinput,1,'(\d*)',updateLog3,'Store every n\'th spam message. 
The same as for non spam but helps prevent spam corpuses being skewed by flooding.<br />
It is recommended that this be set depending on spam volume. Default value = 1, log every message.'],
 [LogRollDays,'Roll the Logfile How Often?',5,textinput,'14','([\d\.]+)',undef,
'ASSP closes and renames the log file after this number of days. Decimals are ok. For example: 14 or 0.5'],

[0,0,0,heading,'Security'],
 [runAsUser,'Run as UID',20,textinput,'nobody','(\S*)',undef,
  'The *nix user name to assume after startup: assp or nobody -- requires ASSP restart.'],
 [runAsGroup,'Run as GID',20,textinput,'nobody','(\S*)',undef,
  'The *nix group to assume after startup: assp or nogroup -- requires ASSP restart.'],
 [ChangeRoot,'Change Root',60,textinput,'','(.*)',undef,
  'Non-blank means to run in chroot jail in *nix. You need an etc/protocols file to make this work<br />
  Copy or link the file to your new root directory -- requires ASSP restart.'],
 # I hate password input, but if you like it, uncomment this line and comment the next one. -- just quit bugging me about it!
 [webAdminPassword,'Web Admin Password',20,passinput,'nospam4me','(.{5,})',undef,
 #[webAdminPassword,'Web Admin Password',20,textinput,'nospam4me','(.{5,})',undef,
  'This is your password for the administrative interface -- Must be at least 5 characters long.'],
 [allowAdminConnectionsFrom,'Allow Admin Connections From*',40,textinput,'','(.*)',ConfigMakeRe,
  'This is an optional list of IP addresses from which you will accept web admin connections, separated by pipes (|).<br />
  For example: 127.0.0.1|10. Blank means accept all connections. 127.0.0.1 means accept connections from the localhost.<br />
  Note that IP source addresses are very easy to spoof, so this should not be considered as a security feature.<br />
  <span class="negative">If you make a mistake here you will disable your web admin interface and have to manually edit your configuration file to fix it.</span>'],

[0,0,0,heading,'Other Settings'],
 [MaxFiles,'Max Files',10,textinput,14009,'(\d+)',undef,
  'If you\'re not using subjects as file names, this is the maximum
 number of files to keep in each collection (spam &amp; nonspam)<br />
 It\'s actually less than this -- files get a random number between 1 and $MaxFiles.'],
 [MaxBytes,'Max Bytes',10,textinput,20000,'(\d+)',undef,
  'How many bytes of the message will ASSP look at? For example: 20000'],
 [RamSaver,'Use less RAM to rebuild the spamdb',0,checkbox,'','(.*)',undef,
  'Checking this slows down your rebuildspamdb process, but will do so with less ram.'],
 [OutgoingBufSize,'Size of TCP/IP Buffer',20,textinput,102400,'(\d+)',undef,
  'If ASSP talks to the internet over a modem change this to 4096, 102400 is the default.
  For example: 102400'],
 [OrderedTieHashSize,'Ordered-Tie hash table size',10,textinput,5000,'(\d+)',undef,
  'Tunable value of the size of the hash tables used by both ASSP and rebuildspamdb.pl (default = 5000).<br />
 Larger numbers mean more RAM, fewer disk hits. Adjust down to use less RAM.'],
 [MaxErrors,'Max Errors',10,textinput,'10','(\d+)',undef,
  'If the smtp destination sends $MaxErrors 501s, 502s, 503s... the connection is dropped.'],
 [RestartEvery,'Restart Every',10,textinput,'0','(\d+)',undef,
  'Program terminates after this many seconds, this is really only useful if ASSP runs in a script that restarts it after
   it stops.<br /> Note: the current timeout must expire before the new setting is loaded.'],
 [UpdateWhitelist,'Save Whitelist',10,textinput,3600,'(\d+)',undef,
  'Save a copy of the white list every this many seconds.<br />
  Note: the current timeout must expire before the new setting is loaded, or you can restart.'],
 [totalizeSpamStats,'Upload Consolidated Spam Statistics',0,checkbox,1,'(.*)',undef,'Checked means your ASSP will upload its totalled statistics to the <a href="http://assp.sourceforge.net/cgi-bin/total.pl" rel="external">ASSP web site totals</a>.<br />
  This is a great marketing tool for the ASSP project; please don\'t disable it unless you\'ve got
  a good reason to.<br /> No private information is being disclosed by this upload.'],
 );

 $MakeRE{ispip}=\&setISPRE;
 $MakeRE{allowAdminConnectionsFrom}=\&setACFRE;
 $MakeRE{acceptAllMail}=\&setAMRE;
 $MakeRE{localDomains}=\&setLDRE;
 $MakeRE{spamaddresses}=\&setSARE;
 $MakeRE{spamLovers}=\&setSLRE;
 $MakeRE{noProcessing}=\&setNPREL;
 $MakeRE{whiteListedDomains}=\&setWLDRE;
 $MakeRE{blackListedDomains}=\&setBLDRE;
 $MakeRE{heloBlacklistIgnore}=\&setHBIRE;
 $MakeRE{LocalAddresses_Flat}=\&setLAFRE;

 # allow override for default web admin port
 if($ARGV[1]=~/^\d+$/) {
  for (@Config) {
   if($_->[0] eq 'webAdminPort' ) {
    $_->[4]=$ARGV[1];
    last;
   }
  }
 }
 open(F,"<$base/assp.cfg"); local $/; (%Config)=split(/:=|\n/,<F>); close F;
 # load defaults
 unless( %Config ) { for $c (@Config) {$Config{$c->[0]}=$c->[4] if $c->[0];} }

 $plusIcon = "http://assp.sourceforge.net/plusIcon.png";
 $minusIcon = "http://assp.sourceforge.net/minusIcon.png";
 $noIcon = "http://assp.sourceforge.net/noIcon.png";

 $NavMenu = '
  <a href="lists"><img src="' . $noIcon . '" alt="noicon" /> <b>Whitelist/Redlist</b></a><br />
  <a href="maillog"><img src="' . $noIcon . '" alt="noicon" /> <b>See the maillog tail</b></a><br />
  <a href="analyze"><img src="' . $noIcon . '" alt="noicon" /> <b>Analyze an email</b></a><br />
  <a href="stats"><img src="' . $noIcon . '" alt="noicon" /> <b>Statistics</b></a><br />
  <a href="donations"><img src="' . $noIcon . '" alt="noicon" /> <b>Donations</b></a><br />';

 $StyleSheet = 'body,p,td {font-size: 11pt; font-family: Verdana, Arial, Helvetica, sans-serif;
	color: #666666; padding: 2px; margin: 0; background-color: #e9e9e9;}
 a {font-weight: bold; color: #666666;}
 a:hover {color: black;}
 img {border-style: none;}  
 hr{color: #6699CC; border-style: none; height: 1px; background-color: #6699CC;}
 h2{margin: 0 0 5px 0; border-width: 0 0 1px 0; border-color: #6699CC; border-style: solid;
  text-align: right; padding-bottom: 3px; color: #6699CC;}
 table {border-color: #de7813; border-width: 0px; border-style: solid;}
 td {padding: 2px 2px 2px 5px;}  
 .navMenu{clear: right; float: left; font-family: arial, serif; font-size: .8em;	display: block;
  padding: 9px 5px 5px 3px;	width: 15em; 
  border: dotted #E6E6E6 1px; margin: 5px 0 0 0; background-color: white;}
 .navMenu a{text-decoration: none; color: #003366; font-weight: normal;}
 .menuLevel1 a{color: #666666; font-weight: bold;}
 .menuLevel2 a{margin-left: 5px; color: #666666; font-weight: bold;} 
 .menuLevel3 a{color: #003366; margin-left: 9px;}
 .content{margin: 5px 0 0 16em; background-color: white; padding: 5px;
  border: dotted 1px; border-color: #E6E6E6; font-size: smaller;}
 .contentHead{padding: 3px;  border: solid 1px #666666; font-weight: bold; 
  text-align: center; color: #CCCCCC; font-weight: bold; background-color: #666666; margin-bottom: 
  1px; margin-right: 4px;}
 .option{background-color: #EEEFF1; padding: 3px; margin: 0 6px 0 0; border-width: 
  0px; border-style: none;}
 .optionTitle {padding: 0 0 5px 5px; border-width: 0px; border-style: none;
  background-color: #EEEFF1; color: #003366;}
 .optionValue {padding: 0 3px 0 5px; border-width: 0px; border-style: none; 
  text-align: justify; font-size: smaller; background-color: #EEEFF1;}
 .note {background-color: white; border: solid 1px #66CC66;
  padding: 5px; font-weight: bold; text-align: justify; font-family: arial, serif; margin: 3px;
  color: #333333;}
 .rightButton{padding: 5px; border-style: none; border-width: 0px; 
  margin: 3px; text-align: right;}
 .statBox{background-color: white; border-color: black; border-width: 1px; padding: 5px; 
  border-style: solid; text-align: left; width: 80%; margin-left: 10%;}	
 .statsOptionTitle{font-weight: normal; color: #333333; font-size: 90%;}
 .textBox {background-color: #e9e9e9; padding: 5px; border: #BFBFBF 1px solid} 
 .smaller {font-size: 50%; font-weight: normal;}
 .sectionHeader {color: #cccccc; font-weight: bold; background-color: #666666;}
 .noBorder {border-width: 0px; border-style: none;}
 .warning {border: solid 1px #9966CC; margin-top: 3px;
 font-weight: bold; padding: 5px;}
 .negative {color: red;}
 .positive {color: green;}
 .shadow {background: url(http://assp.sf.net/dropShadow.gif) right bottom no-repeat; 
  background-color: #EEEFF1;}
 .underline {border: dotted #CCCCCC; border-width: 0 0 1px 0;}';

$JavaScript = "
<script type=\"text/javascript\">
<!--
function toggleDisp(nodeid)
{
  if(nodeid.substr(0,9) == 'setupItem')
    nodeid = nodeid.substr(9);
  layer = document.getElementById('treeElement' + nodeid);
  img = document.getElementById('treeIcon' + nodeid);	
  if(layer.style.display == 'none')
  {
    layer.style.display = 'block';    
    img.src = '$minusIcon';
    if(document.getElementById('setupItem' + nodeid))    
      document.getElementById('setupItem' + nodeid).style.display = 'block';
  }
  else
  {
    layer.style.display = 'none';
    img.src = '$plusIcon';
    if(document.getElementById('setupItem' + nodeid))
      document.getElementById('setupItem' + nodeid).style.display = 'none';
  }
}
function expand(expand, force)
{
  counter = 0;
  while(document.getElementById('treeElement' + counter))
  {
    if(!expand)
    {
      //dont shrink if this element is the one passed in the URL
      arr = document.getElementById('treeElement' + counter).getElementsByTagName('a');
      txt = ''; found = 0;
      loc = new String(document.location);
      for(i=0; i < arr.length; i++)
      {
        txt = txt + arr.item(i).href;
        tmpHref = new String(arr.item(i).href);
        if(tmpHref.substr(tmpHref.indexOf('#')) == loc.substr(loc.indexOf('#')))
        {
          //give this tree node the right icon
          document.getElementById('treeIcon' + counter).src = '$minusIcon';
          found = 1;
        }
      }
      if(!found | force)
      {
        document.getElementById('treeIcon' + counter).src = '$plusIcon';
        document.getElementById('treeElement' + counter).style.display = 'none';
        if(document.getElementById('setupItem' + counter))
          document.getElementById('setupItem' + counter).style.display = 'none';			
      }
    }
    else
    {
      document.getElementById('treeElement' + counter).style.display = 'block';
      document.getElementById('treeIcon' + counter).src = '$minusIcon';
      if(document.getElementById('setupItem' + counter))
        document.getElementById('setupItem' + counter).style.display = 'block';
    }
    counter++;
  }
}
//make the 'rel's work
function externalLinks() 
{
  if (!document.getElementsByTagName) 
    return;
  var anchors = document.getElementsByTagName(\"a\");
  for (var i=0; i<anchors.length; i++)
  {
    var anchor = anchors[i];
    if (anchor.getAttribute(\"href\")
      && anchor.getAttribute(\"rel\") == \"external\")
      anchor.target = \"_blank\";
  }
}
window.onload = externalLinks;
// -->
</script>";

$headers = "HTTP/1.1 200 OK
Content-type: text/html
Cache-control: no-cache

<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"
    \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">

<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\">
<head>
	<meta http-equiv=\"content-type\" content=\"application/xhtml+xml; charset=utf-8\" />
	<title>ASSP (Anti SPAM SMTP Proxy)</title>
<style type=\"text/css\">
<!--
$StyleSheet
//-->
</style>
$JavaScript
</head>
<body>
<p>
  <a href=\"http://assp.sourceforge.net/\"><img src=\"http://assp.sf.net/logo.jpg\" alt=\"ASSP\" /></a>
  <a href=\"http://assp.sourceforge.net/\">$title</a>
</p>
  <div class=\"navMenu\"><div><div style=\"text-align: center;\">
  <a onmousedown=\"expand(1, 1)\">Expand All</a>&nbsp;&nbsp;&nbsp;
  <a onmousedown=\"expand(0, 1)\">Collapse All</a></div>
  <hr />
  <div class=\"menuLevel1\"><a href=\"/\"><img src=\"$plusIcon\" alt=\"plusicon\" /> Main</a><br /></div>";
	
  my $counter = 0;
  foreach $c (@Config) 
  {
    if(@{$c} == 5)
    {
      $headers .= "</div>\n  <div class=\"menuLevel2\">\n  <a onmousedown=\"toggleDisp('$counter')\">" .
        "<img id=\"treeIcon$counter\" src=\"http://assp.sf.net/plusIcon.png\" alt=\"plusicon\" /> " .
        "$c->[4]</a>\n</div>\n<div id=\"treeElement$counter\" style=\"padding-left: 3px; display: block\">";
      $counter++;
    }
    else
    {
      $headers .= "\n    <div class=\"menuLevel3\"><a href=\"/#$c->[0]\">$c->[0]</a></div>";
    }
  }				
  $headers .= "</div>
  <div class=\"menuLevel1\">$NavMenu</div>
  <hr />
  <div style=\"text-align: center;\">
    <a href=\"http://assp.sf.net\">ASSP</a> v$version$modversion<br />
  </div> 
  </div>
<script type=\"text/javascript\">
<!--  
expand(0,0);
// -->
</script>
  ";

$footers = '
<div class="content">
 <a href="http://sourceforge.net">
 <img src="http://sourceforge.net/sflogo.php?group_id=69172&amp;type=1" alt="SourceForge Logo" />
 </a>
</div>';

 fixConfigSettings();
 $SIG{HUP}=\&reloadConfigFile;
 $SIG{PIPE} = "IGNORE";
}

$EmailAdrRe="[^()<>@,;:\\/\"\\[\\]\000-\040]+";
$EmailDomainRe='\w[\w\.\-]*\.\w+';

# Notes on general operation & program structure
# I'm using IO::Select, so don't make any changes that block for long
# as new connections come we create a pair of entries in a hash %Con
# based on the hash of the filehandle, so $Con{$fh} has data for this
# connection. $Con{$fh}->{friend} is the partner socket for the smtp proxy.
# ->{ip} is the ip address of the connecting client
# ->{relayok} tells if we can relay mail for this client
# ->{getline} is a pointer to a function that should be called whan a line of input is received for this filehandle
# ->{mailfrom} is the envelope sender (MAIL FROM: <address>)
# ->{outgoing} is a buffer for outgoing socket traffic (see $writable & &sendque)
# ->{rcpt} are the addresses from RCPT TO: <address> (space separated)
# ->{header} is where the header (and eventually the first 10000 bytes) are stored
# ->{maillog} if present stream logging is enabled
# ->{maillogbuf} buffer for storing unwritten stream log while waiting for isspam decision
# ->{maillogfh} is the filehandle for logging lines to the maillog
# ->{mailloglength} is the length logged so far (we stop after 10000 bytes)
# ->{spamfound} is a flag used to signal if an email is determined to be spam.
# ->{maillength} is the same as mailloglength but is not reset.
#
# After connection the {getline} field functions like a state machine
# redirecting input to subsequent handlers
#
# whitebody -> getline
#   getbody ->
#     error -> (disconnects)
#     getline -> getheader ->
#       whitebody -> getline
#         error -> (disconnects)
#
# getline looks for MAIL FROM, RCPT TO, RSET
# getheader looks for a blank line then tests for whitelist / spamaddresses
# getbody looks for the . and calls isspam, the Bayesian spam test
# whitebody waits for . and redirects client to server
# error waits for . ignoring data from client (and finishes the maillog)
#
# the server has states like this:
#
# skipok -> reply
#
# skipok traps the 250 ok response from the NOOP Connection from
# reply echos server messages to the client
# reply also looks for a 235 AUTH OK and sets {relayok}=1

use IO::Select;
use IO::Socket;

sub serviceCheck {}
sub d {}

eval(q[sub d {
 $time=gmtime(); $time=~s/... (...) (..) (........) ..(..)/$2 $1 $4 $3/;
 print DEBUG "$time <$_[0]>";
 }]) if $DEBUG;

if($AsADaemon) {
 fork() && exit;
 close STDOUT;
 close STDERR;
 $silent=1;
}

if($pidfile) {open(F,">$base/$pidfile"); print F $$; close F;}

sub RemovePid {
 if ($pidfile) {
  d('RemovePid');
  unlink("$base/$pidfile");
 }
}

if($DEBUG) {open(DEBUG, ">$base/".time.".dbg"); binmode(DEBUG); my $oldfh = select(DEBUG); $| = 1; select($oldfh);}
if($logfile && open(LOG,">>$base/$logfile")) {my $oldfh = select(LOG); $| = 1; select($oldfh);}

if($AsAService) {
 eval(<<'EOT');
 use Win32::Daemon;
 mlog(0,'Starting as a Service');
 Win32::Daemon::StartService();
 
 # Wait until the service manager is ready for us to continue...
 while( SERVICE_START_PENDING != Win32::Daemon::State() ) { sleep( 1 ); }
 Win32::Daemon::State( SERVICE_RUNNING );
 
sub serviceCheck {
 d(50);
 if(Win32::Daemon::State() == SERVICE_STOP_PENDING ) {
  d(51);
  mlog(0,'Service Stopping');
  &saveWhitelist;
  Win32::Daemon::State( SERVICE_STOPPED );
  Win32::Daemon::StopService();
  exit;
 }
}

EOT
 print STDERR "error: $@\n" if $@;
 print LOG "error: $@\n" if $@;
}

init();
while(1) {
 &MainLoop;
}

sub init {
 mlog(0,"ASSP version $version$modversion initializing");
 if ($CanUseLDAP) {
  mlog(0,"LDAP module installed and available.");
 } else {
  mlog(0,"LDAP module not installed.");
 }
 if ($CanUseDNS) { 
  mlog(0,"Net DNS module installed and available.");
 } else {
  mlog(0,"Net DNS module not installed.");
 }
 if ($CanUseAddress) {
  mlog(0,"Email address valid module installed and available.");
 } else {
  mlog(0,"Email address valid module not installed.");
 }
 if ($CanUseSPF) {
  mlog(0,"SPF query module installed and available.");
 } elsif ($AvailSPF) {
  mlog(0,"SPF query module installed but Net DNS required.");
 } else {
  mlog(0,"SPF query module not installed.");
 }
 

 $readable = new IO::Select( );
 $writable = new IO::Select( );
 
 $lsn = newListen($listenPort,\&NewSMTPConnection);
 $WebSocket = newListen($webAdminPort,\&NewWebConnection);
 mlog(0,"Listening for mail connections at $listenPort and admin connections at $webAdminPort");
 if($listenPort2) {
  $lsn2 = newListen($listenPort2,\&NewSMTPConnection);
  mlog(0,"Listening for additional mail connections at $listenPort2");
 }
 # handle the possible relayhost / smarthost option
 if($relayHost && $relayPort) {
  $Relay=newListen($relayPort,\&NewSMTPConnection);
  mlog(0,"Listening for relay connections at $relayPort");
 }
 
 $nextNoop=time;
 $endtime=$nextNoop+$RestartEvery;
 $saveWhite=$nextNoop+$UpdateWhitelist;
 
 $SIG{INT}=sub {mlog(0,'Sig INT'); &SaveWhitelist; kill 6,$$ if $ENV{windir}; RemovePid(); exit;};
 $SIG{TERM}=sub {mlog(0,'Sig TERM'); &SaveWhitelist; RemovePid(); exit;};
 
 my ($uid,$gid)=getUidGid($runAsUser,$runAsGroup) if ($runAsUser || $runAsGroup);
 if($ChangeRoot) {
  my $chroot;
  eval('$chroot=chroot($ChangeRoot)');
  if($@) {
   mlog('',"Request to change root to '$ChangeRoot' failed: $@");
   die "Request to change root to '$ChangeRoot' failed: $@";
  } elsif(! $chroot) {
   mlog('',"Request to change root to '$ChangeRoot' did not succeed: $!");
   die "Request to change root to '$ChangeRoot' did not succeed: $!";
  } else {
   $chroot=$ChangeRoot; $chroot=~s/(\W)/\\$1/g;
   $base=~s/^$chroot//i;
   chdir("/");
   mlog('',"Successfully changed root to '$ChangeRoot' -- new base is '$base'");
  }
 }
 switchUsers($uid,$gid) if ($runAsUser || $runAsGroup);

 # create folders if they're missing
 mkdir "$base/$spamlog",0700; mkdir "$base/$notspamlog",0700;
 my $dir=$correctedspam;
 $dir=~s/\/.*?$//;
 mkdir "$base/$dir",0700;
 mkdir "$base/$correctedspam",0700; mkdir "$base/$correctednotspam",0700;
 
 # put this after chroot so the paths don't change
 $SpamdbObject=tie %Spamdb,orderedtie,"$base/$spamdb" if $spamdb;
 mlog(0,"Warning: Bayesian spam database is small or empty: '$base/$spamdb'") if $spamdb && -s "$base/$spamdb" < 10000;
 $HeloBlackObject=tie %HeloBlack,orderedtie,"$base/$spamdb.helo" if $spamdb;
 $DnsblObject=tie %Dnsbl,orderedtie,"$base/$dnsbl" if $dnsbl;
 mlog(0,"Warning: DNS blacklist database is small or empty: '$base/$dnsbl'") if $dnsbl && -s "$base/$dnsbl" < 10000;
 $WhitelistObject=tie %Whitelist,orderedtie,"$base/$whitelistdb";
 mlog(0,"Warning: Whitelist is small or empty: '$base/$whitelistdb' (ignore if this is a new install)") if $whitelistdb && -s "$base/$whitelistdb" < 1000;
 $RedlistObject=tie %Redlist,orderedtie,"$base/$redlistdb";
 $GreylistObject=tie %Greylist,orderedtie,"$base/$greylist" if $greylist;
 if($AvDbs) {
  mlog(0,"Loading Virus definitions ...");
  Av->init( {path => ($AvPath || $base), databases => $AvDbs} );
  mlog(0,"Virus definitions loaded; count=".Av->count);
 }
 $Stats{starttime}=time;
 $Stats{version}=$version;
 $Stats{pid}=$$;
 &ResetStats;
 mlog(0,"Starting");
}

sub newListen {
 my($port,$handler)=@_;
 my ($interface,$p)=$port=~/(.*):(.*)/;
 my $s;
 if($interface) {
  $s = new IO::Socket::INET(Listen => 10, LocalPort => $p, Reuse=>1, LocalAddr => $interface);
 } else {
  $s = new IO::Socket::INET(Listen => 10, LocalPort => $port, Reuse=>1);
 }
 if(! $s) {
  mlog('',"Couldn't create server socket on port '$port' -- maybe another service is running or I'm not root (uid=$>)?");
  return undef;
 }
 $SocketCalls{$s}=$handler;
 $readable->add($s); # add to select list
 $s;
}

sub getUidGid { my ($uname,$gname)=@_;
 return if $AsAService;
 eval('getgrnam(root);getpwnam(root);');
 if($@) {
  # windows pukes "unimplemented" for these -- just skip it
  mlog('',"Warning: uname and/or gname are set ($uname,$gname) but getgrnam / getpwnam give errors: $@");
  return;
 }
 my $gid;
 if($gname) {
  $gid = getgrnam($gname);
  if(defined $gid) {
  } else {
   mlog('',"Could not find gid for group '$gname' -- not switching effective gid -- quitting");
   die "Could not find gid for group '$gname' -- not switching effective gid -- quitting";
  }
 }
 my $uid;
 if($uname) {
  $uid = getpwnam($uname);
  if(defined $uid) {
  } else {
   mlog('',"Could not find uid for user '$uname' -- not switching effective uid -- quitting");
   die "Could not find uid for user '$uname' -- not switching effective uid -- quitting";
  }
 }
 ($uid,$gid);
}

sub switchUsers { my ($uid,$gid)=@_;
 return if $AsAService;
 my($uname,$gname)=($runAsUser,$runAsGroup);
 $>=0;
 if($> != 0) {
  mlog('',"requested to switch to user/group '$uname/$gname' but cannot set effective uid to 0 -- quitting; uid is $>");
  die "requested to switch to user/group '$uname/$gname' but cannot set effective uid to 0 -- quitting; uid is $>\n";
 }
 $<=0;
 if($gid) {
  $)=$gid;
  if($)+0==$gid) {
   mlog('',"Switched effective gid to $gid ($gname)");
  } else {
   mlog('',"Failed to switch effective gid to $gid ($gname) -- effective gid=$) -- quitting");
   die "Failed to switch effective gid to $gid ($gname) -- effective gid=$) -- quitting";
  }
  $(=$gid;
  if($(+0==$gid) {
   mlog('',"Switched real gid to $gid ($gname)");
  } else {
   mlog('',"Failed to switch real gid to $gid ($gname) -- real uid=$(");
  }
 }
 if($uid) {
  # do it both ways so linux and bsd are happy
  $<=$uid; $>=$uid; $<=$uid; $>=$uid;
  if($>==$uid) {
  mlog('',"Switched effective uid to $uid ($uname)");
  } else {
   mlog('',"Failed to switch effective uid to $uid ($uname) -- real uid=$< -- quitting");
   die "Failed to switch effective uid to $uid ($uname) -- real uid=$< -- quitting";
  }
  if($<==$uid) {
   mlog('',"Switched real uid to $uid ($uname)");
  } else {
   mlog('',"Failed to switch real uid to $uid ($uname) -- real uid=$<");
  }
 }
}

sub MainLoop {
 my $wait=7; # keep it short enough for servicecheck to be called regularly
 my ($canread,$canwrite)=IO::Select->select($readable,$writable,undef,$wait);
 foreach $fh (@$canwrite) {
  my $l=length($Con{$fh}->{outgoing});
  print DEBUG "$fh $Con{$fh} l=$l\n" if $DEBUG;
  if(length($Con{$fh}->{outgoing})) {
   my $written=syswrite($fh,$Con{$fh}->{outgoing},$OutgoingBufSize);
   if($DEBUG) {
    if($written < 200) {
     print DEBUG "wrote: ($written)<",substr($Con{$fh}->{outgoing},0,$written),">\n";
    } else {
     print DEBUG "wrote: ($written)<long text>\n";
    }
   }
   $Con{$fh}->{outgoing}=substr($Con{$fh}->{outgoing},$written);
   $l=length($Con{$fh}->{outgoing});
   # test for highwater mark
   if($written>0 && $l < $OutgoingBufSize && $Con{$fh}->{paused}) {
    $Con{$fh}->{paused}=0;
    $readable->add($Con{$fh}->{friend});
   }
  }
  if(length($Con{$fh}->{outgoing})==0) {
   $writable->remove($fh);
  }
 }
 foreach $fh (@$canread) {
  $SocketCalls{$fh}->($fh) if $fh && $SocketCalls{$fh};
 }
 d(6);
 serviceCheck(); # for win32 services
 # timer related issues
 $time=time;
 if($time - $ltime == $wait) {
  # noactive connections -- check for maintenance
  d(8);
  if($time >= $saveWhite) {
   d(9);
   &SaveWhitelist;
   $saveWhite=$time+$UpdateWhitelist;
  }
  if($RestartEvery && $time >= $endtime) {
   # time to quit -- after endtime and we're bored.
   $opencon=(keys %Con);
   if($opencon == 0) {
    &SaveWhitelist;
    mlog(0,"Restarting");
    exit 1;
   }
  }
  uploadStats() if($totalizeSpamStats && $time >= $Stats{nextUpload});
  downloadGrey() if(!$nogreydownload && $greylist && $time >= $NextGreylistDownload);
  if($AvDbs && Av->checkReload()) {
   mlog(0,"Reloading virus database ...");
   Av->loadAll();
   mlog(0,"Virus signature database reloaded; count=".Av->count);
  }
 }
 $ltime=$time;
}
d(11);
# Never reached...(we hope)

sub SaveWhitelist {
 d(35);
 mlog(0,"Saving Whitelist");
 $WhitelistObject->flush() if $WhitelistObject;
 $RedlistObject->flush() if $RedlistObject;
 SaveStats();
}

sub mlog{
 my $fh=shift;
 d(34);
 my $m=localtime();
 $m=~s/^... (...) +(\d+) (\S+) ..(..)/$1-$2-$4 $3/;
if($LogRollDays > 0){
   my $t=int(time()/($LogRollDays*24*3600)); # roll log every 14 days
   if($logfile && $mlogLastT && $t != $mlogLastT && $logfile ne "maillog.log") {
   # roll the log
   my ($sec,$min,$hour,$mday,$mon,$year) = localtime(time);
   $mon++; $year-=100;
   my $mm=sprintf("%02d-%02d-%02d",$year,$mon,$mday);
   $archivelogfile = "$mm.$logfile";
   print LOG "$m: Rolling log file -- archive saved as '$archivelogfile'\n";
   print "$m: Rolling log file -- archive saved as '$archivelogfile'\n" unless $silent;
   close LOG;
   rename("$base/$logfile", "$base/$archivelogfile");
   if(open(LOG,">>$base/$logfile")) {my $oldfh = select(LOG); $| = 1; select($oldfh);}
   print LOG "$m: new log file -- old log file renamed to '$archivelogfile'\n";
   SaveConfig();
  }
 $mlogLastT=$t;
 }
 if($fh && $Con{$fh}) {
  $m.= " $Con{$fh}->{ip} <$Con{$fh}->{mailfrom}>";
  my ($to) = $Con{$fh}->{rcpt}=~/(\S+)/;
  $m.= " to: $to" if $to;
 }
 $m.= " $_[0]\n";
 print $m unless $silent;
 print DEBUG $m if $DEBUG;
 print LOG $m if $logfile;
}

#####################################################################################
#                Socket handlers
sub NewSMTPConnection {
 my $fh=shift;
 my ($client, $server, $destination);
 if($fh==$Relay) {
  # a relay connection -- destination is the relayhost
  d(101);
  $destination=$relayHost;
 } else {
  d(1);
  $destination=$smtpDestination;
 }
 if(!($client=$fh->accept)) {
  d("accept failed: $fh");
  return;
 }
 my $ip=$client->peerhost();
 # ip connection limiting
 if ($maxSMTPipSessions && $ip!~$ISPRE && $ip!~$AMRE) {
  $SMTPSession{$ip}++;
  if ($SMTPSession{$ip} >= $maxSMTPipSessions) {
   $Stats{maxSMTPip}++;
   d("limiting ip: $fh");
   mlog(0,"Limiting $ip Connections") if $SessionLog;
   $SMTPSession{$ip}--;
   return;
  } else {
   $SMTPSession{$client}=$fh;
  }
 }
 $server=new IO::Socket::INET(Proto=>'tcp',PeerAddr=>$destination,Timeout=>2);
 if(! $server) {
  mlog('',"Couldn't create server socket to $destination -- aborting connection");
  if (exists $SMTPSession{$client}) {
   delete $SMTPSession{$client};
   $SMTPSession{$ip}--;
  }
  $client->close();
  return;
 }
 addfh($client,\&getline,$server);
 if($sendNoopInfo) {
  addfh($server,\&skipok,$client);
 } else {
  addfh($server,\&reply,$client);
 }
 $Con{$client}->{ip}=$ip;
 print DEBUG "Connected: $client -- $server\n" if $DEBUG;
 $Stats{connects}++;
 if(ok2Relay($ip) || $fh == $Relay ) {
  $Con{$client}->{relayok}=1;
  print DEBUG "$client relaying ok: $ip\n" if $DEBUG;
 }
 $time=gmtime(); $time=~s/... (...) (..) (........) ..(..)/$2 $1 $4 $3/;
 $Con{$client}->{rcvd}="Received: from $ip ([$ip] helo=) by $myName ; $time -0000\r\n";
 print DEBUG  "\n* connect ip=$Con{$client}->{ip} relay=<$Con{$client}->{relayok}> *\n" if $DEBUG;
 mlog(0, "Connected: $ip:".$client->peerport());
 $Con{$server}->{noop}="NOOP Connection from: $ip, $time -0000 relayed by $myName\r\n" if $sendNoopInfo;
 # overall session limiting
 if ($maxSMTPSessions) {
  $SMTPSession{Total}++;
  $SMTPSession{$client}=$fh;
  if ($SMTPSession{Total}>=$maxSMTPSessions) {
   d("limiting sessions: $fh");
   $readable->remove($lsn) if $lsn;
   $readable->remove($lsn2) if $lsn2;
   $readable->remove($Relay) if $Relay;
   $Stats{maxSMTP}++;
   mlog(0,"Limiting Total Connections") if $SessionLog;
  }
 }
 # check if options files have been updated and need to be re-read
 if(time-$lastOptionCheck > 60) { # check for updates each 60 seconds
  for $f (@PossibleOptionFiles) {
   ${$f}=optionList($Config{$f},$f) if $Config{$f}=~/^ *file: *(.+)/i && fileUpdated($1);
  }
  $lastOptionCheck=time;
 }
}

sub SMTPTraffic {
 my $fh=shift;
 my $buf;
 if($fh->sysread($buf,4096)>0) {
  d(2);
  $this=$Con{$fh};
  $buf=$this->{_}.$buf;
  if((my $sb=$this->{skipbytes})>0) {
   # support for XEXCH50 thankyou Microsoft for making my life miserable
   my $l=length($buf);
   print DEBUG "skipbytes=$sb l=$l -> " if $DEBUG;
   if($l >= $sb) {
    sendque($this->{friend},substr($buf,0,$sb)); # send the binary chunk on to the server
    $buf=substr($buf,$sb);
    delete $this->{skipbytes};
   } else {
    sendque($this->{friend},$buf); # send the binary chunk on to the server
    $this->{skipbytes}=$sb-=length($buf);
    $buf='';
   }
   print DEBUG "skipbytes=$this->{skipbytes}\n" if $DEBUG;
  }
  $bn=$lbn=-1;
  while(($bn=index($buf,"\n",$bn+1)) >= 0) {
   $s=substr($buf,$lbn+1,$bn-$lbn);
   if(defined($this->{bdata})) { $this->{bdata}-=length($s); }
   print DEBUG  "doing <$s>\n" if $DEBUG;
   Maillog($fh,$s) if $Con{$fh}->{maillog};
   $Con{$fh}->{getline}->($fh,$s);
   last unless $Con{$fh};  # it's possible that the connection can be deleted while there's still something in the buffer
   $lbn=$bn;
  }
  d(3);
  if($Con{$fh}) {
   ($this->{_})=substr($buf,$lbn+1);
   if(length($this->{_}) > $MaxBytes) {
    d(4);
    if(defined($this->{bdata})) { $this->{bdata}-=length($this->{_}); }
    Maillog($fh,$this->{_}) if $Con{$fh}->{maillog};
    $Con{$fh}->{getline}->($fh,$this->{_});
    ($this->{_})='';
   }
  }
 } else {
  d(5);
  done($fh);
 }
}

sub check4update {
 # only check every 15 seconds
 my $fil=shift;
 return if $check4updateTime{$fil} + 15 > time;
 $check4updateTime{$fil}=time;
 my @s=stat(${$fil});
 my $mtime=$s[9];
 if ($mtime != $FileUpdate{$fil}) {
  # reload
  $FileUpdate{$fil}=$mtime;
  open(F,"<${$fil}");
  local $/="\n";
  my $l; my %h;
  while($l=<F>) {
   $l=~y/\r\n\t //d;
   next unless $l;
   $h{lc $l}=1;
  }
  close F;
  %{$fil}=%h;
 }
}

sub SetRE { my ($var,$r,$f,$log)=@_;
 eval("\$$var=qr/\$r/$f");
 if($@) {
  mlog(0,"Regular Expression Error in '$r' for $log: $@");
 }
}

sub setAMRE {
 SetRE(AMRE,"^($_[0])",'i','Accept All Mail');
}

sub setISPRE {
 SetRE(ISPRE,"^($_[0])",'i','ISP & Secondary MX');
}

sub ok2Relay {
 my $ip=shift;
 return 1 if $acceptAllMail && $ip=~$AMRE;
 if($relayHostFile) {
  check4update(relayHostFile);
  return 1 if $relayHostFile{$ip};
 }
 return 1 if PopB4SMTP($ip);
 # failed all tests -- return 0
 0;
}

sub PopB4SMTP {
 return 0 unless $PopB4SMTPFile;
 unless ($TriedDBFileUse) {
  eval 'use DB_File';
  mlog(0,"Could not load module DB_File: $@") if $@;
  $TriedDBFileUse=1;
 }
 my $ip=shift;
 my %hash;
 tie %hash,  'DB_File', $PopB4SMTPFile, O_READ, 0400, $DB_HASH;
 if($hash{$ip}) {
  return 1;
 } else {
  return 0;
 }
}

sub setACFRE {
 SetRE(ACFRE,"^($_[0])",'i',"Allow Admin Connections From");
}

sub NewWebConnection {
 my $s=$WebSocket->accept;
 return unless $s;
 if($allowAdminConnectionsFrom) {
  my $ip=$s->peerhost();
  mlog('',"admin connection from $ip");
  if($ip!~$ACFRE) {
   mlog('',"admin connection from $ip rejected by allowAdminConnectionsFrom");
   $s->close();
   return;
  }
 }
 $readable->add($s);
 $SocketCalls{$s}=\&WebTraffic;
}

sub WebTraffic {
 my $fh=shift;
 if($fh->sysread($buf,4096)>0) {
  local $_=$WebCon{$fh}.=$buf;
  if(length($_) > 1030000) {
   # throw away connections longer than 1M to prevent flooding
   WebDone($fh);
   return;
  }
  if(/Content-length: (\d+)/i) {
   my $l=$1;
   webRequest($fh,$1,$2) if	/(.*\n)\r?\n\r?(.*)/s && length($2) >= $l;
  } elsif(/\n\r?\n/) {
   webRequest($fh,$_);
  }
 } else {
  # connection closed
  WebDone($fh);
 }
}

sub WebDone {
 my $fh=shift;
 delete $SocketCalls{$fh};
 delete $WebCon{$fh};
 $readable->remove($fh);
}

# done with a file handle -- close him and his friend(s)
sub done {
 my $fh=shift;
 d(12);
 done2($Con{$fh}->{friend});
 done2($Con{$fh}->{forwardSpam});
 done2($fh);
}

# close a file handle & clean up associated records
sub done2 {
 my $fh=shift;
 d(13);
 return unless $fh;
 my $this=$Con{$fh};
 my $ip=$this->{ip};
 return unless $this;
 print DEBUG  "closing $fh\n" if $DEBUG;
 # close the maillog if it's still open
 my $f=$this->{maillogfh};
 close $f if $f;
 # remove from the select structure
 delete $SocketCalls{$fh};
 $readable->remove($fh);
 $writable->remove($fh);
 # close it
 $fh->close;
 # delete the Connection data
 delete $Con{$fh};
 # delete the Session data & re-add sockets.
 if (exists $SMTPSession{$fh}) {
  delete $SMTPSession{$fh};
  $readable->add($lsn) if $lsn && (!$readable->exists($lsn));
  $readable->add($lsn2) if $lsn2 && (!$readable->exists($lsn2));
  $readable->add($Relay) if $Relay && (!$readable->exists($Relay));
  $SMTPSession{Total}-- if $maxSMTPSessions;
  $SMTPSession{$ip}-- if $maxSMTPipSessions;
 }
}

# adding a socket to the Select structure and Con hash
sub addfh {
 my ($fh,$getline,$friend) =@_;
 d(14);
 $SocketCalls{$fh}=\&SMTPTraffic;
 $readable->add($fh);
 binmode($fh);
 $Con{$fh}={};
 my $this=$Con{$fh};
 $this->{getline}=$getline;
 $this->{friend}=$friend;
}

# sendque enques a string for a socket
sub sendque {
 my ($fh,$message)=@_;
 my $l=length($message);
 print DEBUG "sq: $fh l=$l\n" if $DEBUG;
 return unless $fh && $Con{$fh};
 $writable->add($fh);
 $Con{$fh}->{outgoing}.=$message;
 if(!$Con{$fh}->{paused} && length($Con{$fh}->{outgoing}) > $OutgoingBufSize) {
  $Con{$fh}->{paused}=1;
  print DEBUG "pausing\n" if $DEBUG;
  $readable->remove($Con{$fh}->{friend});
 }
}

#####################################################################################
#                SMTP stuff

# compile the regular expression (RE) for the local domains list (LDRE)
sub setLDRE {
 SetRE(LDRE,"^($_[0])\$","i","Local Domains");
}

# returns true if this address is local
sub localmail {
 my $h=shift;
 $h = $1 if $h=~/\@(.*)/;
 return 1 if $localDomains && $h=~$LDRE;
 if($localDomainsFile) {
  check4update(localDomainsFile);
  return 1 if $localDomainsFile{lc $h};
 }
 0;
}

sub localmailaddress {
 my $h=shift;
 my $retcode;
 my $retmsg;
 $h = $1 if $h=~/\@(.*)/;
 # do LDAP lookup
 $current_email = "$1$h";
 $ldapflt = $LDAPFilter;
 $ldapflt =~ s/EMAILADDRESS/$current_email/g;
 print DEBUG  "doing LDAP lookup with $ldapflt in $LDAPRoot\n" if $DEBUG;
 $ldap = Net::LDAP->new( $LDAPHost );
 if(! $ldap) {
  mlog($fh,"Couldn't contact LDAP server at $LDAPHost -- aborting connection");
  seterror($fh,"451 Could not check recipient, try later\r\n",1);
  return 0;
 }
 # bind to a directory anonymous or with dn and password
 if ($LDAPLogin) {
  $mesg = $ldap->bind($LDAPLogin,
 			    password => $LDAPPassword
                      );
 } else {
  # mlog($fh,"LDAP anonymous bind");
  $mesg = $ldap->bind;
 }
 $retcode = $mesg->code;
 if ($retcode) {
  # $retmsg=$mesg->error_text();
  # mlog($fh,"LDAP bind error: $retcode - Login Problem?");
  mlog($fh,"LDAP bind error: $retcode -- aborting connection");
  seterror($fh,"451 Could not check recipient, try later\r\n",1);
  return 0;
 }
 # perform a search
 $mesg = $ldap->search( base   => $LDAPRoot,
 			      filter => $ldapflt,
 			      attrs => ['cn']
                       );
 $retcode = $mesg->code;
 # mlog($fh,"LDAP search: $retcode");
 if($retcode > 0) {
  mlog($fh,"LDAP search error: $retcode -- aborting connection");
  seterror($fh,"451 Could not check recipient, try later\r\n",1);
  return 0;
 }
 $entry_count = $mesg->count;
 $retmsg = $mesg->entry(1);
 # mlog($fh,"LDAP Results: $entry_count : $retmsg");
 print DEBUG  "got $entry_count result(s) from LDAP lookup\n" if $DEBUG;
 $mesg = $ldap->unbind;  # take down session
 return $entry_count
}

# compile the regular expression for the spam addresses
sub setSARE {
  my (@uad, @u, @d);
  for $a (split(/\|/,$_[0])) {
   if($a=~/\S\@\S/) {
    push(@uad,$a);
   } elsif( $a=~/^\@/ ) {
    push(@d,$a);
   } else {
    push(@u,$a);
   }
  }
  my @s;
  push(@s,'^('.join('|',@uad).')$') if @uad;
  push(@s,'^('.join('|',@u).')@') if @u;
  push(@s,'('.join('|',@d).')$') if @d;
  my $s=join("|",@s);
  $s='<not a valid list>' unless $s;
  SetRE(SARE,$s,'i',"Spam Addresses");
}

# compile the no-processing regular expression
sub setNPREL {
  my (@uad, @u, @d);
  for $a (split(/\|/,$_[0])) {
   if($a=~/\S\@\S/) {
    push(@uad,$a);
   } elsif( $a=~/^\@/ ) {
    push(@d,$a);
   } else {
    push(@u,$a);
   }
  }
  my @s;
  push(@s,'^('.join('|',@uad).')$') if @uad;
  push(@s,'^('.join('|',@u).')@') if @u;
  push(@s,'('.join('|',@d).')$') if @d;
  my $s=join("|",@s);
  $s='<not a valid list>' unless $s;
  SetRE(NPREL,$s,'i',"No Processing");
}

# a line of input has been received from the smtp client
sub getline {
 my($fh,$l)=@_;
 d(15);
 my $this=$Con{$fh};
 my $server=$this->{friend};
 print DEBUG  "gl: <$l>\n" if $DEBUG;
 if($l=~/^ *(helo|ehlo) .*?([^<>,;"'\(\)\s]+)/i) {
  my $helo = $2;
  my $helo2 = $helo;
  $helo=~s/(\W)/\\\1/g;
  $l =~ s/$helo/$myName/ ;
  $this->{helo}=$helo2;
  $this->{rcvd}=~s/=\)/=$helo2\)/;
  $this->{rcvd}=~s/(.{60,75}) /$1\r\n  /g unless $this->{rcvd}=~/\n/; # wrap long lines
 } elsif($l=~/mail from:.*?($EmailAdrRe\@$EmailDomainRe)/io) {
  $this->{rcpt}=$this->{header}=''; # reset everything else
  $this->{mailfrom}=$1;
  $this->{noprocessing}=0;
 } elsif($l=~/rcpt to: *(.*)/i) {
  my $e=$1;
  my ($u,$h);
  #enforce valid email address pattern
  if ($CanUseAddress && $DoRFC822) {
   if ($l=~/rcpt to:\s*<*([^\r\n>]*).*/i) {
    my $RO_e=$1;
    if (!Email::Valid->address($RO_e)) {
     # couldn't understand recipient
     sendque($fh, "553 malformed address: $RO_e\r\n");
     mlog($fh,"malformed address: '$RO_e'");
     $Stats{norelays}++;
     return;
    }
   }
  }
  if($e=~/[\!\%\@]\S*\@/) {
   # blatent attempt at relaying
   sendque($fh, $NoRelaying."\r\n");
   mlog($fh,"relay attempt blocked for (evil): $e");
   $Stats{norelays}++;
   return;
  } elsif($e=~/([a-z\-_\.]+)!([a-z\-_\.]+)$/i) {
   # someone give me one good reason why I should support bang paths! grumble...
   $u="$2@";
   $h=$1;
  } elsif($l=~/rcpt to:.*?($EmailAdrRe\@)($EmailDomainRe)/io) {
   ($u,$h)=($1,$2);
  } elsif($defaultLocalHost && $l=~/rcpt to:.*?<($EmailAdrRe)>/io) {
   ($u,$h)=($1,$defaultLocalHost);
   $u.='@';
  } else {
   # couldn't understand recipient
   sendque($fh, $NoRelaying."\r\n");
   mlog($fh,"relay attempt blocked for (parsing): $e");
   $Stats{norelays}++;
   return;
  }
  $this->{islocal}=localmail($h);
  # skip check when RELAYOK
  if (!$this->{relayok}) {
   # Need Check?
   if($LocalAddresses_Flat || $DoLDAP) {
    $this->{islocalmailaddress}=0;
   }
   # check recipient against flat list?
   if($LocalAddresses_Flat) {
    if("$u$h"=~$LAFRE) {
     $this->{islocalmailaddress}=1;
     print DEBUG  "$u$h validated by flat LocalAddresses list\n" if $DEBUG;
    }
   }
   # Need another check?
   if (!$this->{islocalmailaddress}) {
    # check recipient against LDAP ?
    if ($DoLDAP) {
     if ($CanUseLDAP) {
      $this->{islocalmailaddress}=localmailaddress($h);
     } else {
      $this->{islocalmailaddress}=localmail($h);
      mlog($fh,"Net::LDAP not installed, cannot check: $u$h");
     }
    }
   }
  } else {
   $this->{islocalmailaddress}=localmail($h);
  }
  if(!($this->{relayok}) && (! $this->{islocal} || ($u.$h)=~/\%/) || $u =~/\@\w+/) {
   sendque($fh, $NoRelaying."\r\n");
   mlog($fh,"relay attempt blocked for: $u$h");
   $Stats{norelays}++;
   return;
  }
  if($noProcessing) {
   if("$u$h"=~$NPREL) {
    $this->{noprocessing}|=1;
   } else {
    $this->{noprocessing}|=2;
   }
  }
  if($spamaddresses && $this->{islocal} && "$u$h"=~$SARE) {
   $this->{addressedToSpamBucket}=1;
  }
  if($EmailInterfaceOk && $this->{relayok} && $this->{islocal}) {
   if(lc $u eq lc "$EmailSpam\@") {
    $this->{getline}=\&SpamReport;
    mlog($fh,"Email spamreport");
    sendque($fh,"250 ok\r\n");
    return;
   } elsif(lc $u eq lc "$EmailHam\@") {
    $this->{getline}=\&HamReport;
    mlog($fh,"Email hamreport");
    sendque($fh,"250 ok\r\n");
    return;
   } elsif(lc $u eq lc "$EmailWhitelist\@") {
    $this->{getline}=\&EmailWhite;
    mlog($fh,"Email Whitelist");
    for $a (split(/ /,$this->{rcpt})) {EmailWhiteAdd($a,$this)};
    sendque($fh,"250 ok\r\n");
    return;
   }
  }
  # $this->{rcpt}.="$u$h ";
  # update: accept SpamBucket addresses in every case
  if (($LocalAddresses_Flat || $DoLDAP) && (! $this->{addressedToSpamBucket})) {
   if (($this->{islocalmailaddress}) || ($this->{relayok}) && ! $this->{islocal}) {
    $this->{rcpt}.="$u$h ";
    mlog($fh,"Recipient accepted: $u$h");
   } else {
    if ($NoValidRecipient) {
     $reply = $NoValidRecipient."\r\n";
    } else {
     $reply = "550 5.1.1 User unknown\r\n";
    }
    $reply =~ s/EMAILADDRESS/$u$h/g;
    sendque($fh, $reply);
    mlog($fh,"Invalid address rejected: $u$h");
    # increment error and drop line if necessary
    if($this->{serverErrors}++ > $MaxErrors) {
     mlog($fh,"Max errors ($MaxErrors) exceeded -- dropping connection");
     done($fh);
    }
    return;
   }
  } else {
   #mlog($fh,"Recipient accepted unchecked: $u$h");
   $this->{rcpt}.="$u$h ";
  }
 } elsif( $l=~/^ *XEXCH50 +(\d+)/i ) {
  $this->{skipbytes}=$1;
  print DEBUG "XEXCH50 b=$1\n" if $DEBUG;
 } elsif( $l=~/^ *DATA/i || $l=~/^ *BDAT (\d+)/i ) {
  # drop line if no recipients left
  if ($this->{rcpt} !~/@/) {
    mlog($fh,"No recipients left -- dropping connection");
    seterror($fh,"503 5.5.2 Need Recipient\r\n",1);
    return;
  }
  if($1) {
   $this->{bdata}=$1;
  } else { 
   delete $this->{bdata};
  }
  if($this->{noprocessing}==1 ||
   $noProcessing && $this->{mailfrom}=~$NPREL) {
   # all addresses are on no-processing list
   $this->{noprocessing}=1;
   if($BlockNPExes) {
    $this->{getline}=\&whitebodyNoExe;
   } else {
    $this->{getline}=\&whitebody;
    $Stats{noprocessing}++;
    mlog($fh,"message proxied without processing - (attachments unchecked)");
   }
  } else {
   MaillogStart($fh); # notify the stream logging to start logging
   $this->{getline}=\&getheader;
  }
  if ($this->{rcpt} =~/@/) {
   $Stats{messages}++;
  }
 } elsif( $l=~/^ *RSET/i ) {
  $this->{mailfrom}=$this->{rcpt}=$this->{header}='';
  $this->{noprocessing}=0;
 }
 sendque($server, $l);
}

# compile the blacklisted domains regular expression
sub setBLDRE {
 SetRE(BLDRE1,"($_[0])\$","i","Blacklisted Domains");
 SetRE(BLDRE2,"($_[0]) ",'i',"Blacklisted Domains");
}

# compile the helo-blacklist ignore regular expression
sub setHBIRE {
 SetRE(HBIRE,"^($_[0])\$","i","HELO Blacklisted Ignore");
}

# get the header part of the DATA.
sub getheader {
 my($fh,$l)=@_;
 d(18);
 my $this=$Con{$fh};
 $this->{header}.=$l;
 if($l=~/^\.?[\r\n]*$/) {
  $this->{maillength}=length($this->{header});
  # header is done
  if($npRe && $this->{header}=~$npReRE) { 
   $this->{noprocessing}=1;
   if($BlockNPExes) {
    $this->{getline}=\&whitebodyNoExe;
   } else {
    $Stats{noprocessing}++;
    mlog($fh,"message proxied without processing - (attachments unchecked)");
    isnotspam($fh);
   }
  } elsif(onwhitelist($fh,$this->{header})) {
   if ((!$SPFWL || ($SPFWL && SPFok($fh))) && (!$RBLWL || ($RBLWL && RBLok($fh)))) {   
    if($BlockWLExes) {
     $this->{getline}=\&whitebodyNoExe;
    } else {
     Maillog($fh,'',2); # tell maillog this isn't spam
     $SpamProb=0; addSpamProb($fh);
     isnotspam($fh);
     mlog($fh,'local or whitelisted - (attachments unchecked)');
    }
   }
  } elsif($blackListedDomains && ($this->{mailfrom}=~$BLDRE1 || $this->{senders}=~$BLDRE2)) {
   $Stats{blacklisted}++;
   thisIsSpam($fh,'has blacklisted domain',$blDomainLog,$SpamError,$blTestMode);
  } elsif(!$noHeloBlacklist
    && (! $heloBlacklistIgnore || $this->{helo} !~ $HBIRE)
    && $HeloBlackObject && $HeloBlack{$this->{helo}}) {
   $Stats{helolisted}++;
   thisIsSpam($fh,"has spam helo '$this->{helo}'",$spamHeloLog,$SpamError,$hlTestMode);
  } elsif($this->{addressedToSpamBucket}) {
   $Stats{spambucket}++;
   thisIsSpam($fh,'has spam address',$spamBucketLog,$SpamError,$sbTestMode);
  } elsif(SPFok($fh) && RBLok($fh)) {
   # cleared all the above rules - off to bayesian testing if SPF and RBL is OK.
   $this->{getline}=\&getbody;
  }
 }
}

# do SPF (sender policy framework) checks
sub SPFok {
 my($fh)=@_;
 my $this=$Con{$fh};
 d('SPFok');
 return 1 unless (!$this->{relayok} && $CanUseSPF && $ValidateSPF && $this->{ip}!~$ISPRE);
 my $query = new Mail::SPF::Query (ipv4       => $this->{ip},
					     sender     => $this->{mailfrom},
	    				     helo       => $this->{helo},
	                             trusted    => 1,
	                             guess      => $LocalPolicySPF,
		                       myhostname => $myName,
	                             sanitize   => 1,
		                       debug      => $DebugSPF,
		                       debuglog   => sub { mlog($fh,"Debug: @_") }
	                             );    
 my ($per_result, $smtp_comment, $header_comment);
 foreach my $recip (split(' ', $this->{rcpt})) {
  ($per_result, $smtp_comment, $header_comment) = $query->result2($recip);
  # Keep processing SPF records until all recipients are checked otherwise breakout if fail
  if ($per_result eq 'fail') {last;}
 }
 my $received_spf;
 $received_spf = "Received-SPF: $per_result ($header_comment) client-ip=$this->{ip};";
 $received_spf .= " envelope-from=$this->{mailfrom};" if ( defined( $this->{mailfrom} ) );
 $received_spf .= " helo=$this->{helo};" if ( defined( $this->{helo}) );
 mlog($fh,"$received_spf") if $SPFLog;
 $this->{header}=~s/\n(\r?\n)/\nX-Assp-$received_spf$1$1/i if $AddSPFHeader; 
 if ($per_result eq 'fail') {
  # This email fails SPF rules for the sending domain.  Apply SPF Failure Rules
  $Stats{spffails}++;
  $SpamProb=1; 
  thisIsSpam($fh,'failed SPF checks',$SPFFailLog,"550 5.7.1 $smtp_comment",$spfTestMode);
  return 0;
 } else {
  return 1;
 }
}

# do RBL checks
sub RBLok {
 my($fh)=@_;
 my $this=$Con{$fh};
 d('RBLok');
 return 1 unless (!$this->{relayok} && $CanUseRBL && $ValidateRBL && $this->{ip}!~$ISPRE);
 my $rbl = RBL->new(lists       =>   [@rbllist],
                          server      =>   $nameservers[0],
                          max_hits    =>   $RBLmaxhits,
                          max_replies =>   $RBLmaxreplies,
                          query_txt   =>   1,
                          max_time    =>   $RBLmaxtime,
                          timeout     =>   $RBLsocktime) ;

 my ($received_rbl,$rbl_result,$lookup_return);
 $lookup_return = $rbl->lookup($this->{ip});
 my @listed_by  = $rbl->listed_by();
 $rbls_returned = $#listed_by + 1; 
 if ($rbls_returned >= $RBLmaxhits ) {
  $received_rbl = "Received-RBL: fail ($myName: local policy) rbl=@listed_by; client-ip=$this->{ip};";
  $Stats{rblfails}++;
  $SpamProb=1; 
  thisIsSpam($fh,'failed RBL checks',$RBLFailLog,"550 5.7.1 Blacklisted by @listed_by",$rblTestMode);
  $rbl_result=0;	  
 } elsif ($rbls_returned > 0) {
  $received_rbl = "Received-RBL: neutral ($myName: local policy) rbl=@listed_by; client-ip=$this->{ip};";
  $rbl_result=1;
 } else {
  $received_rbl = "Received-RBL: pass ($myName: local policy) rbl=none; client-ip=$this->{ip};";
  $rbl_result=1;
 }
 $this->{header}=~s/\n(\r?\n)/\nX-Assp-$received_rbl$1$1/i if $AddRBLHeader;
 if ($RBLLog) {
   mlog(0,"Lookup Returned:  $lookup_return");
   mlog($fh, $received_rbl);
 }
 return $rbl_result;
}

# compile the spam-lovers regular expression
sub setSLRE {
 my (@uad, @u, @d);
 for $a (split(/\|/,$_[0])) {
  if($a=~/\S\@\S/) {
   push(@uad,$a);
  } elsif( $a=~/^\@/ ) {
   push(@d,$a);
  } else {
   push(@u,$a);
  }
 }
 my @s;
 push(@s,'^('.join('|',@uad).')$') if @uad;
 push(@s,'^('.join('|',@u).')@') if @u;
 push(@s,'('.join('|',@d).')$') if @d;
 my $s=join("|",@s);
 $s='<not a valid list>' unless $s;
 SetRE(SLRE,$s,'i',"Spam Lovers");
}

# returns true if all the addresses in the space separated list are spam-lovers
sub allLoveSpam {
 my $rcpt=shift;
 my $c=0;
 for (split(' ',$rcpt)) {
  return 0 unless $_=~$SLRE;
  $c++;
 }
 $c;
}

# the message is not spam -- route it to the server
sub isnotspam {
 my $fh=shift;
 d(23);
 my $done=shift;
 my $this=$Con{$fh};
 my $server=$this->{friend};
 my $m=$this->{header};
 sendque($server, $m);
 $this->{header}='';
 if($done) {
  $this->{getline}=\&getline;
 } else {
  $this->{getline}=\&whitebody;
 }
}

# the message is non spam -- just relay it to the server
sub whitebody { my($fh,$l)=@_;
 my $this=$Con{$fh};
 d(25);
 my $server=$this->{friend};
 $this->{maillength}+=length($l);
 my $done=$l=~/^\.[\r\n]*$/ || defined($this->{bdata}) && $this->{bdata}<=0;
 if($done) {
  $this->{getline}=\&getline;
 } else {
  if($this->{maillength}>=$AVBytes) {
   if (!$AVBytes) {
    checkInfection($fh,$l); # Full Scan
   }
  } else {
   checkInfection($fh,$l); # Scan up to AV Bytes.
  }
 }
 sendque($server, $l);
}

# the message is non spam -- check if it is executable
sub whitebodyNoExe { my($fh,$l)=@_;
 my $this=$Con{$fh};
 d(255);
 my $b=$this->{header}.=$l;
 my $done=$l=~/^\.[\r\n]*$/ || defined($this->{bdata}) && $this->{bdata}<=0;
 checkInfection($fh,$l);
 if( $done || length($b) >=$MaxBytes) {
  $this->{maillength}=length($b);
  if ($this->{noprocessing}==1) {
   if($b=~$badattachRE[$BlockNPExes]) {
    $Stats{viri}++;
    Maillog($fh,'',$npAttachLog);
    mlog($fh,'bad attachment (noprocessing)');
    seterror($fh,$AttachmentError,$done);
   } else {
    $Stats{noprocessing}++;
    mlog($fh,'message proxied without processing -(no bad attachments)');
    isnotspam($fh,$done);
   }
  } elsif ($b=~$badattachRE[$BlockWLExes]) {
   $Stats{viri}++;
   Maillog($fh,'',$wlAttachLog);
   mlog($fh,'bad attachment (local/white)'); 
   seterror($fh,$AttachmentError,$done);
  } else {
   $SpamProb=0; addSpamProb($fh); 
   Maillog($fh,'',2); # tell maillog this isn't spam
   mlog($fh,'local or whitelisted - (no bad attachments)');
   isnotspam($fh,$done);
  }
 }
}

# the message may or may not be spam -- get the body and test it.
sub getbody { my($fh,$l)=@_;
 d(27);
 my $this=$Con{$fh};
 my $b=$this->{header}.=$l;
 my $done=$l=~/^\.[\r\n]*$/ || defined($this->{bdata}) && $this->{bdata}<=0;
 print DEBUG "bd='$this->{bdata}'\n" if $DEBUG;
 checkInfection($fh,$l);
 if( $done || length($b) >=$MaxBytes) {
  $this->{maillength}=length($b);
  if($bombRe && $b=~$bombReRE) {
   $Stats{bombs}++;
   Maillog($fh,'',$spamBombLog);
   mlog($fh,"mail bomb");
   seterror($fh,$bombError,$done);
  } elsif($scriptRe && $b=~$scriptReRE) {
   $Stats{scripts}++;
   Maillog($fh,'',$scriptLog);
   mlog($fh,"contains scripting");
   seterror($fh,$scriptError,$done);
  } elsif($BlockExes && $b=~$badattachRE[$BlockExes]) {
   $Stats{viri}++;
   Maillog($fh,'',$extAttachLog);
   mlog($fh,'bad attachment (external)');
   seterror($fh,$AttachmentError,$done);
  } elsif($this->{spamfound}) {
   # Spam is found to be safe, lets pass it on.
   mlog($fh,"Spam determined to be safe, passing on to recipient");
   isnotspam($fh,$done);
  } elsif(isspam($b,$this->{ip})) {
   $Stats{bspams}++;
   thisIsSpam($fh,'Bayesian Spam',$baysSpamLog,$SpamError,$baysTestMode,$done);
  } else {
   # skip saving Bayesian nonspam -- prevents false negs from corrupting corpus
   $Stats{bhams}++;
   Maillog($fh,'',$baysNonSpamLog);
   addSpamProb($fh);
   mlog($fh,"message ok");
   isnotspam($fh,$done);
  }
 }
}

sub checkInfection {
 my ($fh,$l)=@_;
 my $this=$Con{$fh};
 return unless $AvDbs && (!$this->{islocal} || ($Avlocal && $this->{islocal})); 
 my $av;
 unless($av=$Con{$fh}->{av}) {
  $av=$Con{$fh}->{av}=Av->new();
 }
 $l=~s/([a-zA-Z0-9+\/=]{40,}\s*)/base64decode($1)/e;
 my $r;
 my $n=0;
 for(;$n<length($l); $n++) {
  if($r=$av->addchar(substr($l,$n,1))) {
   # this mail is infected
   my $er=$AvError;
   $er=~s/\$infection/$r->[1]/gi;
   mlog($fh,"virus detected '$r->[1]'");
   $Stats{viri}++;
   $Stats{viridetected}++;
   seterror($fh,$er);
  }
 }
}
#This is spam, lets see if its test mode or spamlover.
sub thisIsSpam {
 my ($fh,$reason,$log,$error,$testmode,$done)=@_;
 my $this=$Con{$fh};
 addSpamProb($fh);
 $this->{spamfound}=1; #Set spamfound flag.
 my $slok;
 if($testmode || ($slok=$spamLovers && allLoveSpam($this->{rcpt}))) {
  if($slok) {
   mlog($fh,"Passing if safe because spamlover: $this->{rcpt}, otherwise $reason");
   $Stats{spamlover}++;
  } else {
   mlog($fh,"Passing if safe because testmode, otherwise $reason");
  }
  # pretend it's not spam
  $this->{header}=~s/\n(\r?\n)/\nFrom: sender not supplied/ unless $this->{header}=~/\nFrom:/i; # add From: if missing
  $this->{header}=~s/\n(\r?\n)/\nSubject: $1$1/ unless $this->{header}=~/\nSubject:/i; # add Subject: if missing
  $this->{header}=~s/\nSubject:/\nSubject: $spamSubject/gi if $spamSubject;
  $this->{header}=~s/\n(\r?\n)/\nX-Assp-Spam: YES$1$1/i if $AddSpamHeader;
  #Lets check if its safe to pass if not already done so. 
  if ($done) {
   Maillog($fh,'',$log); # tell maillog what this is.
   mlog($fh,"Spam determined to be safe, passing on to recipient");
   isnotspam($fh,$done);
  } else {
   $this->{getline}=\&getbody;
  }
 } else {
  Maillog($fh,'',$log); # tell maillog what this is.
  mlog($fh,$reason);
  seterror($fh,$error,$done);
 }
}

# reject the email
sub seterror {
 my($fh,$e,$done)=@_;
 d(28);
 my $this=$Con{$fh};
 $this->{error}=$e;
 if($done) {
  error($fh,".\r\n");
 } else {
  $this->{getline}=\&error;
  $this->{header}=''; # free up some memory
 }
 # detatch the friend -- closing connection to server & disregarding message
 done2($this->{friend});
 $this->{friend}=undef;
}

# ignore what's sent & give reason at the end.
sub error { my($fh,$l)=@_;
 d(29);
 my $this=$Con{$fh};
 if( $l=~/^\.[\r\n]*$/ || defined($this->{bdata}) && $this->{bdata}<=0) {
  print $fh $this->{error}."\r\n";
  done2($fh);
 }
}

# filter off the 250 OK noop response and go to reply
sub skipok {
 d(291);
 my ($fh,$l)=@_;
 if($l=~/^250/) {
  $Con{$fh}->{getline}=\&reply;
 } else {
  reply(@_);
 }
}

# messages from the server get relayed to the client
sub reply {
 my($fh,$l)=@_;
 d(30);
 my $this=$Con{$fh};
 return unless $this;
 my $cli=$this->{friend};
 return unless $cli;
 if($l=~/250-CHUNKING/i) {
  # we'll filter off the chunking directive to avoid BDAT problems.
  d('CHUNKING');
  return;
 } elsif($l=~/250-XEXCH50/i) {
  # we'll filter off the XEXCH50 service, as it only causes troubles
  d('XEXCH50');
  return;
 } elsif($l=~/250-.*STARTTLS/i) {
  # we'll filter off the STARTTLS directive to avoid TLS problems.
  d('STARTTLS');
  return;
 } elsif($l=~/^220/) {
  sendque($fh,$this->{noop}) if $this->{noop};
  delete $this->{noop};
 } elsif($l=~/^235/) {
  # check for authentication response
  $Con{$cli}->{relayok}=1;
  print DEBUG  "$Con{$cli}->{ip}: authenticated\n" if $DEBUG;
  mlog($cli,"Authenticated");
 } elsif($l=~/^354/) {
  d(301);
 } elsif($l=~/^50[0-9]/) {
  if($Con{$cli}->{skipbytes}) {
   d("Resetting skipbytes");
   $Con{$cli}->{skipbytes}=0; # if we got a negative response from XEXCH50 then don't skip anything
  }
  if($this->{serverErrors}++ > $MaxErrors) {
   mlog($fh,"Max errors ($MaxErrors) exceeded -- dropping connection");
   done($fh);
  }
 }
 sendque($cli, $l);
}


#####################################################################################
#                Email Interface

# this mail isn't really a mail -- it's a spam report
sub SpamReport {
 my($fh,$l)=@_;
 my $this=$Con{$fh};
 if( $l=~/^ *DATA/i || $l=~/^ *BDAT (\d+)/i ) {
  if($1) {
   $this->{bdata}=$1;
  } else { 
   delete $this->{bdata};
  }
  $this->{getline}=\&SpamReport2;
  sendque($fh,"354 ok\r\n");
  return;
 } elsif( $l=~/^ *RSET/i ) {
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
  return;
 } elsif( $l=~/^ *QUIT/i ) {
  $this->{getline}=\&getline;
  sendque($this->{friend},"QUIT\r\n");
  return;
 } elsif( $l=~/^ *XEXCH50 +(\d+)/i ) {
  print DEBUG "XEXCH50 b=$1\n" if $DEBUG;
  sendque($fh,"504 need to authenticate first\r\n");
  return;
 }
 sendque($fh,"250 ok\r\n");
}

# we're getting the body of a spam report
sub SpamReport2 {
 my ($fh, $l)=@_;
 $Con{$fh}->{header}.=$l if length($Con{$fh}->{header}) < $MaxBytes;
 if($l=~/^\.[\r\n]/ || defined($this->{bdata}) && $this->{bdata}<=0) {
  # we're done -- write the file & clean up
  my $this=$Con{$fh};
  my $sub=ReportMail($this->{header},$correctedspam);
  $this->{header}='';
  ReturnMail($this->{mailfrom},"$base/spamreport.txt",$sub) unless $NoHaiku;
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
 }
}

# this mail isn't really a mail -- it's a false-positive report
sub HamReport {
 my($fh,$l)=@_;
 my $this=$Con{$fh};
 if( $l=~/^ *DATA/i || $l=~/^ *BDAT (\d+)/i ) {
  if($1) {
   $this->{bdata}=$1;
  } else {
   delete $this->{bdata};
  }
  $this->{getline}=\&HamReport2;
  sendque($fh,"354 ok\r\n");
  return;
 } elsif( $l=~/^ *RSET/i ) {
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
  return;
 } elsif( $l=~/^ *QUIT/i ) {
  $this->{getline}=\&getline;
  sendque($this->{friend},"QUIT\r\n");
  return;
 } elsif( $l=~/^ *XEXCH50 +(\d+)/i ) {
  print DEBUG "XEXCH50 b=$1\n" if $DEBUG;
  sendque($fh,"504 need to authenticate first\r\n");
  return;
 }
 sendque($fh,"250 ok\r\n");
}

# we're getting the body of a ham report
sub HamReport2 {
 my ($fh, $l)=@_;
 $Con{$fh}->{header}.=$l if length($Con{$fh}->{header}) < $MaxBytes;
 if($l=~/^\.[\r\n]/ || defined($this->{bdata}) && $this->{bdata}<=0) {
  # we're done -- write the file & clean up
  my $this=$Con{$fh};
  my $sub=ReportMail($this->{header},$correctednotspam);
  $this->{header}='';
  ReturnMail($this->{mailfrom},"$base/notspamreport.txt",$sub) unless $NoHaiku;
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
 }
}

sub ReportMail {
 my ($bod,$path)=@_;
 my ($sub)=$bod=~/Subject: (.*)/i;
 $sub=~s/\r//;
 $bod=~s/^.*?\n\r?\n\s*//s;
 # remove the spam subject header addition if present
 my $spamsub=$spamSubject;
 if($spamsub) {
  $spamsub=~s/(\W)/\\$1/g;
  $bod=~s/Subject: $spamsub */Subject: /gi;
 }
 $bod=~s/X-Assp-Spam-Prob: .*\n//gi;
 if($bod=~/\nReceived: /) {
  $bod=~s/^.*?\nReceived: /Received: /s;
 } else {
  $bod=~s/^.*?\n((\w[^\n]*\n)*Subject:)/$1/si;
  $bod=~s/\n> /\n/g;
 }
 my $f;
 do {
  $f=int(rand()*100000000);
 } while(-e "$base/$path/$f.rpt");
 open(F,">$base/$path/$f.rpt"); binmode F;
 print F $bod;
 close F;
 $sub;
}

# we're receiving an email to add addresses to the whitelist
sub EmailWhite {
 my($fh,$l)=@_;
 d('EmailWhite');
 my $this=$Con{$fh};
 if( $l=~/^ *DATA/i || $l=~/^ *BDAT (\d+)/i ) {
  if($1) {
   $this->{bdata}=$1;
  } else { delete $this->{bdata}; }
  sendque($this->{friend},"RSET\r\n"); # make sure to reset the pending email
  $this->{getline}=\&EmailWhiteBody;
  sendque($fh,"354 OK Send whitelist body\r\n");
  return;
 } elsif( $l=~/^ *RSET/i ) {
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
  return;
 } elsif( $l=~/^ *QUIT/i ) {
  $this->{getline}=\&getline;
  sendque($this->{friend},"QUIT\r\n");
  return;
 } elsif( $l=~/^ *XEXCH50 +(\d+)/i ) {
  print DEBUG "XEXCH50 b=$1\n" if $DEBUG;
  sendque($fh,"504 need to authenticate first\r\n");
  return;
 } else {
  for $a ($l=~/($EmailAdrRe\@$EmailDomainRe)/og) {
   EmailWhiteAdd($a,$this);
   $this->{rcpt}.="$a ";
  }
 }
 sendque($fh,"250 OK\r\n");
}

sub EmailWhiteBody {
 my($fh,$l)=@_;
 d('EmailWhiteBody');
 my $this=$Con{$fh};
 if($l=~/^\.[\r\n]/ || defined($this->{bdata}) && $this->{bdata}<=0) {
  # mail summary report
  ReturnMail($this->{mailfrom},"$base/whitereport.txt",'',"$this->{rcpt}\n\n$this->{report}\n") unless $NoHaiku;
  delete $this->{report};
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
 } elsif($l=~/message-id:/i) {
  # ignore
 } else {
  for $a ($l=~/($EmailAdrRe\@$EmailDomainRe)/go) {
   EmailWhiteAdd($a,$this);
  }
 }
}

sub EmailWhiteAdd {
 my ($a,$this)=@_;
 my $t=time;
 if(localmail($a)) {
  $this->{report}.="$a: Cannot add local users to whitelist\n";
 } elsif( $Whitelist{lc $a} ) {
  $this->{report}.="$a: already on whitelist\n";
  mlog(0,"email whitelist addition: $a");
  $Whitelist{lc $a}=$t;
 } else {
  $this->{report}.="$a: added\n";
  mlog(0,"email new whitelist addition: $a");
  $Whitelist{lc $a}=$t;
 }
}

sub ReturnMail {
 my($from,$file,$sub,$bod)=@_;
 my $s=new IO::Socket::INET(Proto=>'tcp',PeerAddr=>$smtpDestination,Timeout=>2);
 if(! $s) {
  mlog('',"Couldn't create server socket to $smtpDestination -- aborting ReturnMail connection");
  return;
 }
 addfh($s,\&RMhelo);
 my $this=$Con{$s};
 $this->{to}=$from;
 $this->{from}=$EmailFrom;
 open(F,"<$file") || mlog(0,"Couldn't open '$file' for mail report");
 local $/="\n";
 my $subject=<F>;
 $subject=~s/\s*(.*)\s*/$1 $sub/;
 $this->{subject}=$subject;
 undef $/;
 $this->{body}=<F>.$bod;
 close F;
 $this->{body}=~s/\r?\n/\r\n/g;
}
sub RMhelo { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"helo Expected 220, got: $l");
 } elsif($l=~/^ *220 /) {
  sendque($fh,"HELO $myName\r\n");
  $Con{$fh}->{getline}=\&RMfrom;
 }
}
sub RMfrom { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"from Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  sendque($fh,"MAIL FROM: ".($Con{$fh}->{from}=~/(<[^<>]+>)/?$1:$Con{$fh}->{from})."\r\n");
  $Con{$fh}->{getline}=\&RMrcpt;
 }
}
sub RMrcpt { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"rcpt Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  sendque($fh,"RCPT TO: <$Con{$fh}->{to}>\r\n");
  $Con{$fh}->{getline}=\&RMdata;
 }
}
sub RMdata { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"data Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  sendque($fh,"DATA\r\n");
  $Con{$fh}->{getline}=\&RMdata2;
 }
}
sub RMdata2 { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"data2 Expected 354, got: $l");
 } elsif($l=~/^ *354 /) {
  my $date=gmtime();
  $date=~s/(\w+) +(\w+) +(\d+) +(\S+) +(\d+)/$1, $3 $2 $5 $4/;
  my $this=$Con{$fh};
  sendque($fh,<<EOT);
From: $this->{from}\r
To: $this->{to}\r
Subject: $this->{subject}\r
X-Assp-Report: YES\r
Date: $date +0000\r
\r
$this->{body}\r
.\r
EOT
  $Con{$fh}->{getline}=\&RMdone;
 }
}
sub RMdone { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"done Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  done2($fh); # close and delete
 }
}
sub RMabort {mlog(0,"RMabort: $_[1]"); done2($_[0]);}

#####################################################################################
#                SPAM Detection

# check if the message is spam, based on Bayesian factors in $Spamdb
sub isspam {
 my $msg=$_[0];
 return $SpamProb=0 if $whiteRe && $msg=~$whiteReRE;
 return $SpamProb=1 if $blackRe && $msg=~$blackReRE;
 local $b=clean($_[0]);
 return $SpamProb=0 if $whiteRe && $b=~$whiteReRE;
 return $SpamProb=1 if $WhiteOnly || $blackRe && $b=~$blackReRE;
 my $ip=$_[1];
 d(31);
 my $ip3=$ip;
 $ip3=~s/(\d+\.\d+\.\d+).*/$1/;
 my ($v,$lt,$llt,$t,%seen);
 my @t;
 if(defined($Dnsbl{$ip}) || defined($Dnsbl{$ip3})) {
  mlog('',"$ip dnsbl hit");
  push(@t,0.97);
 }
 if($greylist) {
  if ($ispip && $ip=~$ISPRE) {
   if ($ispgreyvalue) {
    $v=$ispgreyvalue;
   } else {
    $v=$Greylist{x};
   }
  } else {
   $v=$Greylist{$ip3} || $Greylist{x};
  }
  print DEBUG "gl=$v <$Greylist{$ip3}>\n" if $DEBUG;
  push(@t,$v,$v) if $v;
 }
 while($b=~/([-\$A-Za-z0-9\'\.!\240-\377]+)/g) {
  next if length($1) > 20;
  $llt=$lt; $lt=$t; $t=lc $1;
  $t=~s/[,.']+$//; $t=~s/!!!+/!!/g; $t=~s/--+/-/g;
  my $j="$lt $t";
  next if $seen{$j}++ >1; # first two occurances are significant
  push(@t,$v) if ($v=$Spamdb{$j});
 }
 @t=sort {abs($b-.5)<=>abs($a-.5)} @t;
 @t=@t[0..30];
 my $p1=1; my $p2=1; for $p (@t) {if($p) {$p1*=$p; $p2*=(1-$p);}}
 $p1=$p1/($p1+$p2);
 $SpamProb=$p1;
 $p1<.6? 0: 1;
}

# attach a header line to the message if the config option is set
sub addSpamProb {
 return unless $AddSpamProbHeader;
 my $fh=shift;
 my $this=$Con{$fh};
 return if $NoExternalSpamProb && $this->{relayok};
 my $spamprob=sprintf("X-Assp-Spam-Prob: %.5f\r\n",$SpamProb);
 $spamprob = $spamprob."X-Assp-Envelope-From: $this->{mailfrom}\r\n" if (defined($this->{mailfrom}));
 $this->{spamprob}=$spamprob."X-Intended-For: $this->{rcpt}\r\n" if $sendAllSpam;
 $this->{header}=~s/\r?\nX-Assp-Spam-Prob: [0-9\.]+//g; # clear out existing spam prob
 $this->{header}=~s/^(.*?)(\r?\n)\r?\n/$1$2$spamprob$2/s;
}

sub setWLDRE {
 SetRE(WLDRE,"($_[0])\$",'i',"Whitelisted Domains");
}

sub setLAFRE {
  my (@uad, @u, @d);
  for $a (split(/\|/,$_[0])) {
   if($a=~/\S\@\S/) {
    push(@uad,$a);
   } elsif( $a=~/^\@/ ) {
    push(@d,$a);
   } else {
    push(@u,$a);
   }
  }
  my @s;
  push(@s,'^('.join('|',@uad).')$') if @uad;
  push(@s,'^('.join('|',@u).')@') if @u;
  push(@s,'('.join('|',@d).')$') if @d;
  my $s=join("|",@s);
  $s='<not a valid list>' unless $s;
  SetRE(LAFRE,$s,'i',"LocalAddresses");
}

# see if the address in the mailfrom is on the whitelist meanwhile update the whitelist if that seems appropriate
sub onwhitelist {
 my($fh,$header,$a)=@_;
 d(32);
 my $this=$Con{$fh};
 $a=lc $this->{mailfrom};
 my $whitelisted=$this->{relayok};
 $Stats{locals}++ if $whitelisted;
 return $whitelisted unless $a; # don't add to the whitelist unless there's a valid envelope -- prevent bounced mail from adding to the whitelist
 $this->{red}= ( $Redlist{$a} || ($redRe && $header=~$redReRE) );
 my %senders;
 unless($whitelisted) {
  $senders{$a}=1;
  if(! $NotGreedyWhitelist) {
   while($header=~/\n(from|sender|reply-to|errors-to|list-\w+):.*?($EmailAdrRe\@$EmailDomainRe)/igo) {
    $senders{lc $2}=1;
   }
  }
  for $a (keys %senders) {
   return 0 if $a && $Redlist{$a};
   next if localmail($a) || $a eq '';
   if($whiteListedDomains && $a=~$WLDRE) {
    d("wld '$&'");
    $whitelisted=1;
    last;
   } elsif($Whitelist{$a}) {
    d("on whitelist");
    $whitelisted=1;
    last;
   }
  }
  $this->{senders}=join(' ',keys %senders)." "; # used for finding blacklisted domains
  $Stats{whites}++ if $whitelisted;
 }
 # don't add to whitelist if sender is redlisted
 return $whitelisted if $this->{red} || $WhitelistLocalOnly && !$this->{relayok};
 if($whitelisted) {
  # keep the whitelist up-to-date
  my %a=%senders;
  my $t=time;
  $a{$a}=1;
  $header=~s/\n\s+/ /g;
  while($header=~/\n(to|cc): (.*)/ig) {
   my $a=$2;
   while($a=~/($EmailAdrRe\@$EmailDomainRe)/go) {
    $a{lc $1}=1;
   }
  }
  for $a (split(' ',lc $this->{rcpt})) {
   $a{$a}=1;
  }
  for $a (keys %a) {
   next if localmail($a) || ! $a;
   next if $whiteListedDomains && $a=~$WLDRE;
   mlog($fh,"whitelist addition: $a") unless $Whitelist{$a};
   $Whitelist{$a}=$t;
  }
  return 1;
 }
 0;
}

# clean up source email
sub clean {
 local $_="\n".shift;
 my ($helo)=/helo=([^)]+)\)/i;
 $helo=~s/(\w+)/ hlo $1 /g if length($helo) > 19; # if the helo string is long, break it up
 my $rcpt="rcpt ".join(" rcpt ",/($EmailAdrRe\@$EmailDomainRe)/g);
 # replace &#ddd encoding
 s/&#(\d{1,3});?/chr($1)/ge;
 # replace base64 encoding
 s/\n([a-zA-Z0-9+\/=]{40,}\r?\n[a-zA-Z0-9+\/=\r\n]+)/base64decode($1)/gse;
 # clean up quoted-printable references
 s/(Subject: .*)=\r?\n/$1\n/;
 s/=\r?\n//g;
 s/=([0-9a-fA-F]{2})/pack("C",hex($1))/gei;
 s/%([0-9a-fA-F][0-9a-fA-F])/pack('C',hex($1))/ge; # replace url encoding
 # strip out mime continuation
 s/.*---=_NextPart_.*\n//g;
 # mark the subject
 s/\nsubject: (.*)/fixsub($1)/ige;
 # remove received lines
 s/\n(received|Content-Type): .*(\n[\t ].*)*//ig;
 # remove other header lines
 s/(\n[a-zA-Z\-]{2,40}: .*(\n[\t ].*)*){2,}//g;
 # clean up &nbsp; and &amp;
 s/&nbsp;?/ /gi; s/&amp;?/and/gi;
 s/(\d),(\d)/$1$2/g;
 s/\r//g; s/ *\n/\n/g;
 s/\n\n\n\n\n+/\nblines blines\n/g;
 # clean up html stuff
 s/<script.*?>\s*(<!\S*)?/ jscripttag jscripttag /ig;
 while(s/(\w+)(<[^>]*>)((<[^>]*>)*\w+)/$2$1$3/g){} # move html out of words
 s/<([biu]|strong)>/ boldifytext boldifytext /gi;
 # remove some tags that are not informative
 s/<\/?(p|br|div|t[dr])[^>]*>/\n/gi; s/<\/([biu]|font|strong)>//gi;
 s/<\/?(html|meta|head|body|span|o)[^>]*>//ig;
 s/(<a\s[^>]*>)(.*?)(<\s*\/a\s*>)/$1.fixlinktext($2).$3/igse;
 s/<\s*\/a\s*>//gi;
 # treat titles like subjects
 s/<title[^>]*>(.*?)<\/title>/fixsub($1)/ige;
 # remove style sheets
 s/<style[^>]*>.*?<\/style>//igs;
 # remove html comments
 s/<!.*?-->//gs; s/<![^>]*>//g;
 # look for random words
 s/[ a-z0-9][ghjklmnpqrstvwxz_]{2}[bcdfghjklmnpqrstvwxz_0-9]{3}\S*/ randword randword /gi;
 # remove mime separators
 s/\n--.*randword.*//g;
 # look for linked images
 s/(<a[^>]*>[^<]*<img)/ linkedimage linkedimage $1/gis;
 s/<[^>]*href\s*=\s*("[^"]*"|\S*)/fixhref($1)/isge;
 s/http:\/\/(\S*)/fixhref($1)/isge;
 s/(\S+\@\S*\.\w{2,3})\b/fixhref($1)/ge;
 "helo: $helo\n$rcpt\n$_";
}

sub fixhref { my $t=shift; $t=~s/(\w+)/ href $1 /g; $t;}

sub fixlinktext { my $t=shift; $t=~s/(\w+)/atxt $1/g; $t;}

sub fixurl {
 my $a=shift;
 $a=~s/%([0-9a-fA-F][0-9a-fA-F])/pack('C',hex($1))/ge;
 $a;
}

sub fixsub {
 my $s=shift;
 $s=~s/ {3,}/ lotsaspaces /g;
 $s=~s/(\S+)/ssub $1/g;
 "\n$s ssub";
}

sub base64decode {
 my $str = shift;
 my $res;
 $str =~ tr|A-Za-z0-9+/||cd;
 $str =~ tr|A-Za-z0-9+/| -_|;
 while ($str =~ /(.{1,60})/gs) {
  my $len = chr(32 + length($1)*3/4);
 $res .= unpack("u", $len . $1 );
 }
 $res;
}

sub downloadGrey {
 # let's check if we really need to
 my @s=stat("$base/$greylist");
 my $mtime=$s[9];
 if(time - $mtime< 12*3600) {
  $NextGreylistDownload=$mtime + 12*3600;
  return;
 }
 my $peeraddress,$connect;
 if ($proxyserver) {
  mlog(0,"Freshening greylist via Proxy:$proxyserver");
  $peeraddress = $proxyserver;
  $connect = "GET http://assp.sourceforge.net/greylist.txt HTTP/1.0

";
 } else {
  mlog(0,"Freshening greylist via Direct Connection");
  $peeraddress = "assp.sourceforge.net:80";
  $connect = "GET /greylist.txt HTTP/1.1
Host: assp.sourceforge.net

";
 }
 my $s=new IO::Socket::INET(Proto=>'tcp',PeerAddr=>$peeraddress,Timeout=>2);
 if($s) {
  print $s $connect;
  $SocketCalls{$s}=\&GreyTraffic;
  $readable->add($s);
  $NextGreylistDownload=time+3600*12;
  open(GREYTEMP,">$base/$greylist.tmp");
  binmode(GREYTEMP);
  $GreyListLen=0;
 } else {
  mlog(0,"Unable to connect to Greylist Server");
  $NextGreylistDownload=time+3600;
 }
}

sub GreyTraffic {
 my $fh=shift;
 my $buf;
 if($fh->sysread($buf,4096)>0) {
  if(!$GreyListLen) {
   # look for http header
   if($buf=~/content-length: (\d+)/i) {
    $GreyListLen=$1;
   }
   if($buf=~/(.*?)\n\r?\n(.*)/s) {
    $buf=$2;
   }
  }
  if($GreyListLen > 0) {
   print GREYTEMP $buf;
   $GreyListLen-=length($buf);
   if(! $GreyListLen) {
    close GREYTEMP;
    unlink("$base/$greylist");
    rename("$base/$greylist.tmp","$base/$greylist");
    mlog(0,"Greylist download complete.");
    if($GreylistObject) {
     $GreylistObject->resetCache();
    } else {
     $GreylistObject=tie %Greylist,orderedtie,"$base/$greylist" if $greylist;
    }
    $readable->remove($fh);
    $fh->close;
   }
  }
 } else {
  # greylist download interrupted
  $readable->remove($fh);
  $fh->close;
  close(GREYTEMP);
  mlog(0,"Greylist download interrupted");
 }
}

sub uploadStats {
 my $peeraddress,$connect;
 if ($proxyserver) {
  mlog(0,"Uploading stats via Proxy:$proxyserver");
  $peeraddress = $proxyserver;
  $connect = "POST http://assp.sourceforge.net/cgi-bin/upload.pl HTTP/1.0";
 } else {
  mlog(0,"Uploading stats via Direct Connection");
  $peeraddress = "assp.sourceforge.net:80";
  $connect = "POST /cgi-bin/upload.pl HTTP/1.1
Host: assp.sourceforge.net";
 }
 my $s=new IO::Socket::INET(Proto=>'tcp',PeerAddr=>$peeraddress,Timeout=>2);
 if($s) {
  $Stats{timenow}=time;
  $Stats{testmode}=$TestMode;
  my $content=join("\001",%Stats);
  my $len=length($content);
  $connect.="
Content-Type: application/x-www-form-urlencoded
Content-Length: $len

$content";
  print $s $connect;
  $s->close;
 } else {
  mlog(0,"Unable to connect to stats server");
 }
 $Stats{nextUpload}=time+3600*8;
}

sub ResetStats {
 $Stats{nextUpload}=time+3600*8;
 $Stats{locals}=0;
 $Stats{whites}=0;
 $Stats{messages}=0;
 $Stats{noprocessing}=0;
 $Stats{maxSMTP}=0;
 $Stats{maxSMTPip}=0;
 $Stats{spams}=0;
 $Stats{blacklisted}=0;
 $Stats{helolisted}=0;
 $Stats{spambucket}=0;
 $Stats{bhams}=0;
 $Stats{bspams}=0;
 $Stats{viri}=0;
 $Stats{viridetected}=0;
 $Stats{connects}=0;
 $Stats{spamlover}=0;
 $Stats{bombs}=0;
 $Stats{scripts}=0;
 $Stats{spffails}=0;
 $Stats{rblfails}=0;
 open(F,"<$base/asspstats.sav");
 (%OldStats)=split(/\001/,<F>);
 close F;
}

sub SaveStats {
 my %s=%OldStats;
 for(keys %Stats) {$s{$_}+=$Stats{$_}}
 $s{starttime}=$OldStats{starttime} || $Stats{starttime};
 open(F,">$base/asspstats.sav");
 print F join("\001",%s);
 close F;
 %s;
}

#####################################################################################
#                Maillog functions


# find an appropriate name for a maillog file
sub maillogFilename {my $isspam=shift;
 my @dirs=($notspamlog,$spamlog,$incomingOkMail,$viruslog);
 my $maillog=$dirs[$isspam];
 d(19);
 if($UseSubjectsAsMaillogNames) {
  my($sub)=$_[0]=~/Subject: (.*)/;
  $sub=~y/a-zA-Z0-9/_/cs;
  $sub=substr($sub,0,50);
  $sub.="--".(++$Counter);
  return "$base/$maillog/$sub$maillogExt";
 } else {
  my $fn=(time + $Counter++ ) % $MaxFiles;
  "$base/$maillog/$fn$maillogExt";
 }
}

# integrated mail collection subroutine
sub MaillogStart {
 d(361);
 $Con{$_[0]}->{maillog}=1 unless $NoMaillog;
 $Con{$_[0]}->{maillogbuf}=$Con{$_[0]}->{header}=$Con{$_[0]}->{rcvd};
}
sub Maillog {
 my ($fh,$text,$parm)=@_;
 # parm = 2 -- not spam, 3 -- is spam, 4 -- mail ok
 # 5 -- virii, 6 -- ignore, 7 -- ignore && cc to spamaccount
 return unless $Con{$fh}->{maillog};
 my $skipLog=0;
 if ($parm == 2 || $parm == 3) {
  $logCount[$parm]++;
  $skipLog=1 if ($logFreq[$parm]>$logCount[$parm]);
 } 
 if($parm == 4 && !$incomingOkMail || $parm == 5 && !$viruslog || $parm >= 6 || $skipLog) {
  d(364);
  delete $Con{$fh}->{maillogbuf};
  delete $Con{$fh}->{maillog};
  close $Con{$fh}->{maillogfh} if $Con{$fh}->{maillogfh};
  delete $Con{$fh}->{maillogfh};
  delete $Con{$fh}->{mailloglength};
 } elsif($parm > 1) {
  d(362);
  $logCount[$parm]=0;
  # we now know if it is spam or not -- open the file
  $text=$Con{$fh}->{maillogbuf}.$text;
  delete $Con{$fh}->{maillogbuf};
  my $fn=maillogFilename($parm-2,$text);
  $FH='FHaaaaa' unless $FH;
  if(open(++$FH,">$fn")) {
   $Con{$fh}->{maillogfh}=$FH;
   $Con{$fh}->{mailloglength}=0;
   binmode $FH;
  } else {
   mlog($fh, "error opening maillog '$fn': $!");
  }
  # start sending the message to sendAllSpam if appropriate
  $Con{$fh}->{forwardSpam}=forwardSpam($Con{$fh}->{mailfrom},$sendAllSpam,$fh) if (($parm==3 || $parm==7) && $sendAllSpam);
 }
 if(my $h=$Con{$fh}->{maillogfh}) {
  print $h $text;
  if(($Con{$fh}->{mailloglength}+=length($text))>$MaxBytes || $text=~/(^|[\r\n])\.[\r\n]/) {
   d(366);
   close $h;
   delete $Con{$fh}->{maillog} unless $Con{$fh}->{forwardSpam};
   delete $Con{$fh}->{maillogfh};
   delete $Con{$fh}->{mailloglength};
  }
 } elsif(length($Con{$fh}->{maillogbuf})<$MaxBytes) {
  $Con{$fh}->{maillogbuf}.=$text;
 }
 if($Con{$fh}->{forwardSpam}) {
  my $t=$Con{$Con{$fh}->{forwardSpam}};
  if($t->{ready}) {
   sendque($Con{$fh}->{forwardSpam},$text);
  } else {
   $t->{body}.=$text;
  }
 }
}

sub forwardSpam {
 my ($from,$to,$oldfh)=@_;
 my $s=new IO::Socket::INET(Proto=>'tcp',PeerAddr=>$smtpDestination,Timeout=>2);
 if(! $s) {
  mlog('',"Couldn't create server socket to $smtpDestination -- aborting sendAllSpam connection");
  return;
 }
 addfh($s,\&FShelo);
 my $this=$Con{$s};
 $this->{to}=$to;
 $this->{from}=$from;
 $this->{spamprob}=$Con{$fh}->{spamprob};
 $s;
}
sub FShelo { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  FSabort($fh,"helo Expected 220, got: $l");
 } elsif($l=~/^ *220 /) {
  sendque($fh,"HELO $myName\r\n");
  $Con{$fh}->{getline}=\&FSfrom;
 }
}
sub FSfrom { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  FSabort($fh,"from Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  sendque($fh,"MAIL FROM: <$Con{$fh}->{from}>\r\n");
  $Con{$fh}->{getline}=\&FSrcpt;
 }
}
sub FSrcpt { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  FSabort($fh,"rcpt Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  sendque($fh,"RCPT TO: <$Con{$fh}->{to}>\r\n");
  $Con{$fh}->{getline}=\&FSdata;
 }
}
sub FSdata { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  FSabort($fh,"data Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  sendque($fh,"DATA\r\n");
  $Con{$fh}->{getline}=\&FSdata2;
 }
}
sub FSdata2 { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  FSabort($fh,"data2 Expected 354, got: $l");
 } elsif($l=~/^ *354 /) {
  my $this=$Con{$fh};
  $this->{body}=~s/^(.*?)(\r?\n)\r?\n/$1$2$this->{spamprob}$2/s;
  sendque($fh,$this->{body}) if $this->{body};
  $this->{ready}=1;
  $this->{body}='';
  $Con{$fh}->{getline}=\&FSdone;
 }
}
sub FSdone { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  FSabort($fh,"done Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  done2($fh); # close and delete
 }
}
sub FSabort {mlog(0,"FSabort: $_[1]"); done2($_[0]);}

#####################################################################################
#                Web Configuration functions

sub webRequest {
 my ($fh,$head,$data)=@_;

 %webRequests=(
  '/stats' => \&ConfigStats,
  '/lists' => \&ConfigLists,
  '/analyze' => \&ConfigAnalyze,
  '/maillog' => \&ConfigMaillog,
  '/donations' => \&Donations,
 );

 ($page,$j,$qs)=/^\w+ ([^\? ]+)(\?(\S*))?/i;
 $qs=$data if defined $data;
 ($auth)=/Authorization: Basic (\S+)/i;
 ($user,$pass)=split(':',base64decode($auth));
 $qs=~y/+/ /;
 (%qs)=split(/[=&]/,$qs);
 for $k (keys %qs) {$qs{$k}=~s/%([0-9a-fA-F][0-9a-fA-F])/pack(C,hex($1))/ge}
 if($pass eq $webAdminPassword) {
  ConfigQuit($fh) if($page eq '/quit');
  print $fh ((defined ($v=$webRequests{$page}))? $v->(): webConfig());
 } else {
  print $fh "HTTP/1.1 401 Unauthorized
WWW-Authenticate: Basic realm=\"Anti-Spam SMTP Proxy (ASSP) Configuration\"
Content-type: text/html

<html><body><h1>Unauthorized</h1>
</body></html>\n";
 }

 WebDone($fh);
}

sub ConfigQuit {
 mlog('',"Quit requested from admin interface");
 &SaveWhitelist;
 my $fh=shift;
 print $fh "HTTP/1.1 200 OK
Content-type: text/html

<html><body><h1>ASSP Terminated.</h1>
</body></html>
";
 exit;
}

sub ConfigStats {
 #For legacy support, Stats{spams} are effectively non bayesian spams.
 $Stats{spams}= $Stats{blacklisted} + $Stats{helolisted} + $Stats{spambucket}
              + $Stats{spffails} + $Stats{rblfails}
              + $Stats{bombs} + $Stats{scripts}
              + $Stats{viri} + $Stats{viridetected};
 my %os=SaveStats();
 my $s;
 my $uptime=(time - $Stats{starttime})/(24*3600);
 my $uptime2=(time - $os{starttime})/(24*3600);
 my $mpd=sprintf("%.1f",$Stats{messages}/$uptime);
 my $mpd2=sprintf("%.1f",$os{messages}/$uptime2);
 my $spams=$Stats{spams}+$Stats{bspams};
 my $spams2=$os{spams}+$os{bspams};
 my $hams=$Stats{locals} + $Stats{whites} + $Stats{bhams} + $Stats{noprocessing};
 my $hams2=$os{locals} + $os{whites} + $os{bhams} + $Stats{noprocessing};
 my $pct=sprintf("%.1f",$spams/($Stats{messages}-$Stats{locals}+1)*100);
 my $pct2=sprintf("%.1f",$spams2/($os{messages}-$os{locals}+1)*100);
 my $uptime=sprintf("%.3f",$uptime);
 my $uptime2=sprintf("%.3f",$uptime2);
 my $tconns=$Stats{connects};
 if ($Stats{connects}==0) { $tconns=1000000 ; }
 my $tconns2=$os{connects};
 if ($os{connects}==0) { $tconns2=1000000 ; }
 my $rejpct=sprintf("%.1f%%",($Stats{norelays}/$tconns)*100);
 my $rejpct2=sprintf("%.1f%%",($os{norelays}/$tconns2)*100);

$title = 'ASSP Statistics';
$headers . "<div class=\"content\"><h2>$title</h2>
<table class=\"statBox\">
<tr><td colspan=\"3\" class=\"sectionHeader\">Connection Statistics</td></tr>
<tr><td class=\"statsOptionTitle\"><b>ASSP Proxy Uptime:</b></td>
<td class=\"optionValue\">$uptime days</td>
<td class=\"optionValue\">$uptime2 days</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>SMTP Connections Received:</b></td>
<td class=\"optionValue\">$Stats{connects}</td>
<td class=\"optionValue\">$os{connects}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>SMTP Connection Limits - Total:</b></td>
<td class=\"optionValue\">$Stats{maxSMTP}</td>
<td class=\"optionValue\">$os{maxSMTP}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>SMTP Connection Limits - by IP:</b></td>
<td class=\"optionValue\">$Stats{maxSMTPip}</td>
<td class=\"optionValue\">$os{maxSMTPip}</td>
</tr>
<tr><td colspan=\"3\" class=\"sectionHeader\">Totalled Message Statistics</td></tr>
<tr><td class=\"statsOptionTitle\"><b>Messages Processed:</b></td>
<td class=\"optionValue\">$Stats{messages} ($mpd per day)</td>
<td class=\"optionValue\">$os{messages} ($mpd2 per day)</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Relay Attempts Rejected:</b></td>
<td class=\"optionValue\">$Stats{norelays} ($rejpct)</td>
<td class=\"optionValue\">$os{norelays} ($rejpct2)</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Total Messages Blocked:</b></td>
<td class=\"optionValue\">$spams</td>
<td class=\"optionValue\">$spams2</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Total Messages Passed:</b></td>
<td class=\"optionValue\">$hams</td>
<td class=\"optionValue\">$hams2</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Non-Local Mail Blocked:</b></td>
<td class=\"optionValue\">$pct%</td>
<td class=\"optionValue\">$pct2%</td>
</tr>
<tr><td colspan=\"3\" class=\"sectionHeader\">Blocked Message Statistics</td></tr>
<tr><td class=\"statsOptionTitle\"><b>Bayesian Spams:</b></td>
<td class=\"optionValue\">$Stats{bspams}</td>
<td class=\"optionValue\">$os{bspams}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Blacklisted Messages:</b></td>
<td class=\"optionValue\">$Stats{blacklisted}</td>
<td class=\"optionValue\">$os{blacklisted}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>HELO Listed Messages:</b></td>
<td class=\"optionValue\">$Stats{helolisted}</td>
<td class=\"optionValue\">$os{helolisted}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Spam Bucketed Messages:</b></td>
<td class=\"optionValue\">$Stats{spambucket}</td>
<td class=\"optionValue\">$os{spambucket}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Attachments Blocked:</b></td>
<td class=\"optionValue\">$Stats{viri}</td>
<td class=\"optionValue\">$os{viri}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Viruses Detected:</b></td>
<td class=\"optionValue\">$Stats{viridetected}</td>
<td class=\"optionValue\">$os{viridetected}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Spam Bombs Blocked:</b></td>
<td class=\"optionValue\">$Stats{bombs}</td>
<td class=\"optionValue\">$os{bombs}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Scripts Blocked:</b></td>
<td class=\"optionValue\">$Stats{scripts}</td>
<td class=\"optionValue\">$os{scripts}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>SPF Failures:</b></td>
<td class=\"optionValue\">$Stats{spffails}</td>
<td class=\"optionValue\">$os{spffails}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>RBL Failures:</b></td>
<td class=\"optionValue\">$Stats{rblfails}</td>
<td class=\"optionValue\">$os{rblfails}</td>
</tr>
<tr><td colspan=\"3\" class=\"sectionHeader\">Passed Message Statistics</td></tr>
<tr><td class=\"statsOptionTitle\"><b>Bayesian Hams:</b></td>
<td class=\"optionValue\">$Stats{bhams}</td>
<td class=\"optionValue\">$os{bhams}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Whitelisted Messages:</b></td>
<td class=\"optionValue\">$Stats{whites}</td>
<td class=\"optionValue\">$os{whites}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Local Messages:</b></td>
<td class=\"optionValue\">$Stats{locals}</td>
<td class=\"optionValue\">$os{locals}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Noprocessing Messages:</b></td>
<td class=\"optionValue\">$Stats{noprocessing}</td>
<td class=\"optionValue\">$os{noprocessing}</td>
</tr>
<tr><td class=\"statsOptionTitle\"><b>Spamlover Spams Passed:</b></td>
<td class=\"optionValue\">$Stats{spamlover}</td>
<td class=\"optionValue\">$os{spamlover}</td>
</tr>
</table>
</div>
$footers
</body></html>
";
}

sub ConfigLists {
 my $s;
 my $act=$qs{action};
 if($act) {
  my $color=$qs{list} eq 'red'? 'Red': 'White';
  my $list=$color."list";
  for ($qs{addresses}=~/($EmailAdrRe\@$EmailDomainRe)/go) {
   $s.="<div class=\"text\">$_ ";
   $_=lc $_;
   if($act eq 'v') {
    if($list->{$_}) {
     $s.="${color}listed";
    } else {
     $s.="<span class=\"negative\">NOT ${color}listed</span>";
    }
   } elsif($act eq 'a') {
    if($list->{$_}) {
     $s.="<span class=\"positive\">already ${color}listed</span>";
    } else {
     if($color eq 'White' && localmail($_)) {
      $s.="<span class=\"negative\">local addresses not allowed on whitelist</span>";
     } else {
      $s.="added";
      $list->{$_}=time;
      mlog(0,"${color}list addition: $_ (admin)");
     }
    }
   } elsif($act eq 'r') {
    if($list->{$_}) {
     $s.="removed";
     delete $list->{$_};
     mlog(0,"${color}list deletion: $_ (admin)");
    } else {
     $s.="not ${color}listed";
    }
   }
   $s.="</div>\n";
  }
 }
 if($qs{B1}=~/^Show (.)/) {
  local $/="\n";
  if($1 eq 'R') {
   $RedlistObject->flush() if $RedlistObject;
   open(F,"<$base/$redlistdb");
   $s.='<div class="textbox"><b>RedList</b></div>';
  } else {
   $WhitelistObject->flush() if $WhitelistObject;
   open(F,"<$base/$whitelistdb");
   $s.='<div class="textbox"><b>Whitelist</b></div>';
  }
  my $l;
  while($l=<F>) {
   my ($a)=$l=~/([^\002]*)/;
   $s.="<div class=\"textbox\">$a</div>";
  }
  close F;
 }
<<EOT;
$headers
<div class="content">
<h2>Update or Verify the Whitelist/Redlist</h2>
$s
<br />
<form method="post" action=\"\">    
    <table class="textBox" style="width: 99%;">
        <tr>
            <td class="noBorder">Do you want to work with the:
            </td>
            <td class="noBorder">
            <input type="radio" name="list" value="white" checked="checked" /> Whitelist or<br />
            <input type="radio" name="list" value="red" /> Redlist
            </td>
        </tr>
        <tr>
            <td class="noBorder">Do you want to: </td>
            <td class="noBorder"><input type="radio" name="action" value="a" />add<br />
            <input type="radio" name="action" value="r" />remove<br />
            <input type="radio" checked="checked" name="action" value="v" />or verify</td>
            <td class="noBorder">List the addresses in this box:<br />
            <p><textarea name="addresses" rows="3" cols="40">$qs{addresses}</textarea></p>
            </td>
        </tr>
        <tr>
            <td class="noBorder">&nbsp;</td>
            <td class="noBorder"><input type="submit" name="B1" value="Submit" /></td>
            <td class="noBorder">&nbsp;</td>
        </tr>
    </table>
</form>
<div class="textBox">
<p>Post less than 1 megabyte of data at a time.</p>
Note: The redlist is not a blacklist. The redlist is a list of addresses that cannot
contribute to the whitelist, and who are not considered local, even if their mail is
from a local computer. For example, if someone goes on a vacation and turns on their
email's autoresponder, put them on the redlist until they return. Then as they reply
to every spam they receive they won't corrupt your non-spam collection or whitelist.

  <form action="" method="post">
  <table style="width: 90%; margin-left: 5%;">
    <tr>
      <td align="center" class="noBorder"><input type="submit" name="B1" value="Show Whitelist" /></td>
      <td align="center" class="noBorder"><input type="submit" name="B1" value="Show Redlist" /></td>
    </tr>
  </table>
  </form>
      <p class="warning">Warning: If your whitelist or redlist is long, pushing these buttons
      is ill-advised. Use these for testing and while your whitelist is short.</p>
</div>
</div>
$footers
</body></html>
EOT
}

sub ConfigAnalyze {
 my ($s,$st,$wl,%wl);
 my $mail=$qs{mail};
 if($mail) {
  my $name=$myName; $name=~s/(\W)/\\$1/g;
  my ($ip)=$mail=~/Received: from ([0-9\.]+).*?by\s+$name/is;
  my ($ip3)=$ip=~/(.*)\.\d+$/;
  my ($helo)=$mail=~/helo=(.*?)\)/i;
  $wl.="<p></p><div class=\"textBox\">";
  while($mail=~/($EmailAdrRe\@$EmailDomainRe)/go) {
   my $a=lc $1;
   next if $wl{$a}++;
   if( $Redlist{$a} ){
    $wl.="<b>$1 is redlisted<br /></b>\n"
   } else {
    $wl.="<b>$1 is whitelisted<br /></b>\n" if $Whitelist{$a};
   }
  }
  $wl.="<p>" if $wl;
  $wl.="<b>Mail matches White RE: '$&'</b><p>\n" if $whiteRe && $mail=~$whiteReRE;
  $wl.="<b>Mail matches Black RE: '$&'</b><p>\n" if $blackRe && $mail=~$blackReRE;
  $wl.="<b>Mail matches No Processing RE: '$&'</b><p>\n" if $npRe && $mail=~$npReRE;
  $wl.="<b>Mail matches Red RE: '$&'</b><p>\n" if $redRe && $mail=~$redReRE;
  $wl.="<b>Mail matches Mail Bomb RE: '$&'</b><p>\n" if $bombRe && $mail=~$bombReRE;
  $wl.="<b>Mail matches Script RE: '$&'</b><p>\n" if $scriptRe && $mail=~$scriptReRE;
  $wl.="<b>Mail matches helo blacklist '$helo'</b><p>\n" if $HeloBlack{lc $helo};

  $mail=clean(substr($mail,0,10000));
  $wl.="<b>Mail matches White RE: '$&'</b><p>\n" if $whiteRe && $mail=~$whiteReRE;
  $wl.="<b>Mail matches Black RE: '$&'</b><p>\n" if $blackRe && $mail=~$blackReRE;

  my @t; my %got;

  if(defined($Dnsbl{$ip}) || defined($Dnsbl{$ip3})) {
   $wl.="<b>$ip dnsbl hit</b> (adds 0.97 0.97)<p>\n";
   push(@t,0.97);
   push(@t,0.97);
  }
  if($greylist) {
    my $v;
    if ($ispip && $ip=~$ISPRE) {
       if ($ispgreyvalue) {
        $v=$ispgreyvalue;
      } else {
        $v=$Greylist{x};
      }
    } else {
      $v=$Greylist{$ip3} || $Greylist{x};
    }
   $wl.="<b>$ip3 has a greylist value of $v</b> (adds $v $v)<p>\n";
   push(@t,$v,$v) if $v;
  }

  my ($v,$lt,$llt,$t,%seen);
  while($mail=~/([-\$A-Za-z0-9\'\.\,!]+)/g) {
   next if length($1) > 20;
   $llt=$lt; $lt=$t; $t=lc $1;
   $t=~s/[,.']+$//; $t=~s/!!!+/!!/g; $t=~s/--+/-/g;
   my $j="$lt $t";
   next if $seen{$j}++>1;
   push(@t,$v) if ($v=$Spamdb{$j}); $got{$j}=$v if $v;
  }
  my $cnt=0;
  $s.="<tr><td style=\"text-align: right;\">Bad Words</td>
  <td style=\"text-align: center;\">Bad Prob</td>
  <td style=\"text-align: right;\">Good Words</td>
  <td style=\"text-align: center;\">Good Prob</td></tr>\n";
  for (sort {abs($got{$b}-.5)<=>abs($got{$a}-.5)} keys %got) {
   my $g=sprintf("%.4f",$got{$_});
   if($g < 0.5) {
    $s.="<tr><td></td><td></td><td align=\"right\">$_</td><td>$g</td></tr>\n";
   } else {
    $s.="<tr><td align=\"right\">$_</td><td>$g</td><td></td><td></td></tr>\n";
   }
   last if $cnt++ > 20;
  }
  @t=sort {abs($b-.5)<=>abs($a-.5)} @t;
  @t=@t[0..30];
  $st="<br /><p>Analysis totals: ";
  for (@t) {$st.=sprintf ("%.4f ",$_) if $_;}
  $st.="</p>\n";
  my $p1=1; my $p2=1; for $p (@t) {if($p) {$p1*=$p; $p2*=(1-$p);}}
  $p1=$p1/($p1+$p2);
  $st.=sprintf("<p>spam-prob = %.5f</p>\n",$p1);
  $st.="</div>";  
  $mail=~s/([^\n]{70,84}[^\w\n<\@])/$1\n/g; $mail=~s/\s*\n+/\n/g; $mail=~s/<\/textarea>/\/textarea/ig;
 }
<<EOT;
$headers
<div class="content">
<h2>ASSP Mail Analyzer</h2>
<div class="note">This page will show you how ASSP analyzes an email to come up with the assigned spam
probability. You can also see how it pre-processes mail.
</div>
<br />
$wl<table><tr><td>$s</td></tr></table>$st
<br />
<form action="" method="post">
    <table class="textBox"  style="width: 99%;">
        <tr>
            <td class="noBorder" align="center">Copy and paste the mail header and
            body here:<br />
            <textarea name="mail" rows="10" cols="75">$mail</textarea>            
            </td>
        </tr>
        <tr>
            <td class="noBorder" align="center"><input type="submit" name="B1"
            value="Analyze" /></td>
        </tr>
    </table>
</form>
    <p class="note" >Note: Analysis is performed using the current spam database --
if yours was rebuilt since the time the mail was received you'll
receive a different result.</p>
<div class="textbox">
<p>To use this form using <i>Outlook Express</i> do the following. Right-click on the message
of interest. Select <i>Properties</i>. Click the <i>Details</i> tab. Click the <i>message
source</i> button. Right-click on the message source and click <i>Select All</i>. Right-click
again and click <i>Copy</i>. Click on the text box above and paste (Ctrl-V perhaps). Click
the <i>Analyze</i> button.</p>
<p>The page will update to show you the following: if any of the email's addresses are in
the redlist or whitelist, the most and least spammy phrases together with their spaminess,
the resulting probabilities (probabilities may repeat one time), and the final spam probability
score.</p>
</div>
</div>
$footers
</body></html>
EOT
}

sub ConfigMaillog {
 open(F,"<$base/$logfile");
 seek(F,-10000,2) || seek(F,0,0);
 local $/;
 my $s=<F>;
 close F;
 $s=~s/&/&amp;/g;
 $s=~s/</&lt;/g;
 $s=~s/>/&gt;/g;
<<EOT;
$headers
<div class="content">
<h2>ASSP Maillog Tail</h2>
<div class="note">Press your browser's refresh to update this screen. Newest entries are at the end.</div>
<pre>    
    $s
</pre>
</div>
$footers
</body></html>
EOT
}

sub webConfig {  
 $title = 'ASSP Configuration';
 my $r = $headers . '<div class="content"><h2>' . $title . '</h2><form action="" method="post"><div>';
 $ConfigChanged=0;
 undef %qs unless $qs{theButton}; # don't post partial data if somebody's browser's busted
	$counter = 0;
	for $c (@Config) 
	{
		if(@{$c} == 5)
		{
			@tmp = @{$c};
			push(@tmp, "setupItem$counter");
			$r .= $c->[3]->(@tmp);		
			$counter++;
		}
		else{$r .=$c->[3]->(@{$c});}
	} 
 if($ConfigChanged) {
  SaveConfig();
 }
 my $quit="<form method=\"post\" action=\"quit\"><p>Panic button:
 <input type=\"submit\" value=\"Terminate ASSP now!\" /></p></form>" unless $AsAService;

 $r.= "</div>
<div class=\"rightButton\">
  <input name=\"theButton\" type=\"submit\" value=\"Apply Changes\" />
</div>
<div class=\"note\">
  Note: Items marked with an asterisk (*) accept a list separated by | or you can specify a file to read
  the list from this way: file:c:\\assp\\filename.txt <br />
  Files should have one entry per line; anything on a line following a space and numbersign ( #) is ignored (a comment).<br />
  Whitespace at the beginning or end of the line is ignored.
</div>
</form>
</div>
<div class=\"content\">
 <a href=\"http://sourceforge.net\" rel=\"external\">
 <img src=\"http://sourceforge.net/sflogo.php?group_id=69172&amp;type=1\" alt=\"SourceForge Logo\" />
 </a>
 <div class =\"rightButton\" style=\"padding: 0px;\">$quit</div>
</div>
<script type=\"text/javascript\">
<!--
 expand(0, 0);
 string = new String(document.location);
  string = string.substr(string.indexOf('#')+1);
  if(document.forms[0].length)
  {    
    for(i = 0; i < document.forms[0].elements.length; i++)
    {
      if(string == document.forms[0].elements[i].name)
      {
        document.forms[0].elements[i].focus();        
      }
    }
  }
// -->
</script>
</body></html>\n";
 $r;
}

sub Donations {
$headers . "<div class=\"content\">
<h2>ASSP Donations</h2>
<div class=\"note\">ASSP is here thanks to the following people, please feel free to donate to support the ASSP project.</div>
<table style=\"width: 99%; \" class=\"textBox\">
<tr>
<td class=\"underline\">John Hanna the founder and developer of ASSP up to version 1.0.12</td>
<td class=\"underline\"><a href=\"https://www.paypal.com/xclick/business=johnhanna77%40yahoo.com&amp;item_name=Support+ASSP&amp;item_number=assp&amp;no_note=1&amp;tax=0&amp;currency_code=USD\" rel=\"external\">Donate</a></td>
</tr>
<tr>
<td class=\"underline\">AJ the designer behind ASSP's web interface &amp; site.</td>
<td class=\"underline\">&nbsp;</td>
</tr>
<tr>
<td class=\"underline\">John Calvi the developer of ASSP from version 1.0.12.</td>
<td class=\"underline\"><a href=\"https://www.paypal.com/xclick/business=jcalvi%40lewis.com.au&amp;item_name=Support+ASSP&amp;item_number=assp&amp;no_note=1&amp;tax=0&amp;currency_code=USD\" rel=\"external\">Donate</a></td>
</tr>
<tr>
<td class=\"underline\">Robert Orso the developer of ASSP's LDAP functions.</td>
<td class=\"underline\"><a href=\"https://www.paypal.com/xclick/business=ro%40astronomie.at&amp;item_name=Support+ASSP&amp;item_number=assp&amp;no_note=1&amp;tax=0&amp;currency_code=USD\" rel=\"external\">Donate</a></td>
</tr>
<tr>
<td>&nbsp;</td>
<td></td>
</tr>
<tr>
<td colspan=\"2\">
Special thanks go to......<br />
&nbsp;&nbsp;  Nigel Barling for his contributions to SPF &amp; RBL.<br />
&nbsp;&nbsp;  Mark Pizzolato for his contributions to SMTP Session Limits.<br />
</td>
</tr>
</table>
</div>
<div class=\"content\">
	<a href=\"http://sourceforge.net\" rel=\"external\">
	<img src=\"http://sourceforge.net/sflogo.php?group_id=69172&amp;type=1\" alt=\"SourceForge Logo\" />
	</a>
</div>
</body></html>";
}

sub SaveConfig {
 rename("$base/assp.cfg.bak","$base/assp.cfg.bak.bak");
 rename("$base/assp.cfg","$base/assp.cfg.bak");
 open(F,">$base/assp.cfg");
 for(sort keys %Config) {print F "$_:=$Config{$_}\n";}
 close F;
}

sub textinput {my ($name,$nicename,$size,$func,$default,$valid,$onchange,$description)=@_;
 my $Error=checkUpdate($name,$valid,$onchange);
"<a name=\"$name\"></a>
 <div class=\"shadow\">
 <div class=\"option\">
  <div class=\"optionTitle\">$nicename</div>
  <div class=\"optionValue\">
   <input name=\"$name\" size=\"$size\" value=\"$Config{$name}\" /><br />\n$Error$description\n
  </div>
 </div>  
 &nbsp;
 </div>"; 
}
# everybody wants this, but I hate it -- use it if you care.
sub passinput {my ($name,$nicename,$size,$func,$default,$valid,$onchange,$description)=@_;
 my $Error=checkUpdate($name,$valid,$onchange);
"<a name=\"$name\"></a>
 <div class=\"shadow\">
 <div class=\"option\">
  <div class=\"optionTitle\">$nicename</div>
  <div class=\"optionValue\"><input type=PASSWORD name=\"$name\" size=\"$size\" value=\"$Config{$name}\" /><br />\n$Error$description
  </div>
 </div>  
 &nbsp;
 </div>"; 
}
sub checkbox {my ($name,$nicename,$size,$func,$default,$valid,$onchange,$description)=@_;
 my $Error=checkUpdate($name,$valid,$onchange);
 my $checked=$Config{$name}?'checked="checked"':'';
 "<a name=\"$name\"></a>
 <div class=\"shadow\">
 <div class=\"option\">
  <div class=\"optionTitle\">
   <input type=\"checkbox\" name=\"$name\" value=\"1\" $checked />$nicename<br /></div>
  <div class=\"optionValue\">\n$Error$description
  </div> 
 </div>
 &nbsp;
 </div>";  
}
sub heading {my ($description,$nodeId)=@_[4,5,6];
 "</div>
<div onmousedown=\"toggleDisp('$nodeId')\" class=\"contentHead\">
 $description
</div>
<div id=\"$nodeId\">\n";
}

sub checkUpdate {
 my ($name,$valid,$onchange)=@_;
 return '' unless %qs;
 if($qs{$name} ne $Config{$name}) {
  if($qs{$name}=~/$valid/i) {
   my $new=$1; my $info;
   my $old=$Config{$name};
   $Config{$name}=$new;
   if($onchange) {
    $info=$onchange->($name,$old,$new);
   } else {
    mlog(0,"Admin update: $name changed from '$old' to '$new'");
    ${$name}=$new;
    # -- this sets the variable name with the same name as the config key to the new value
    # -- for example $Config{myName}="ASSP-nospam" -> $myName="ASSP-nospam";
   }
   $ConfigChanged=1;
   return "<span class=\"positive\"><b>*** Updated $info</b></span><br />";
  } else {
   return "<span class=\"negative\"><b>*** Invalid: '$qs{$name}'</b></span><br />";
  }
 }
}

sub fixConfigSettings {
 # This function is called on startup to clean up some settings
 # Primarily these are settings that might be absent from assp.cfg after an upgrade
 $Config{base}=$base;
 $Config{OrderedTieHashSize}=5000 unless $Config{OrderedTieHashSize};
 $Config{ExtensionsToBlock}='exe|scr|pif|vb[es]|js|jse|ws[fh]|sh[sb]|lnk|bat|cmd|com|ht[ab]' unless $Config{ExtensionsToBlock};
 $Config{ispip}=$Config{acceptAllMail} unless $Config{ispip};
 $Config{bombError}='500 Your message was rejected because it appears to be part of a spam bomb -- rephrase your message and try sending it again.' unless $Config{bombError};
 $Config{scriptError}='500 Your email contains html scripting code -- please resend as plain text.' unless $Config{scriptError};
 $Config{wlAttachLog}=5 unless $Config{wlAttachLog};
 $Config{extAttachLog}=5 unless $Config{extAttachLog};
 $Config{spamBombLog}=6 unless $Config{spamBombLog};
 $Config{scriptLog}=3 unless $Config{scriptLog};
 $Config{baysNonSpamLog}=4 unless $Config{baysNonSpamLog};
 $Config{spamHeloLog}=3 unless $Config{spamHeloLog};
 $Config{blDomainLog}=3 unless $Config{blDomainLog};
 $Config{spamBucketLog}=6 unless $Config{spamBucketLog};
 $Config{baysSpamLog}=3 unless $Config{baysSpamLog};
 $Config{SPFFailLog}=3 unless $Config{SPFFailLog};
 $Config{RBLFailLog}=3 unless $Config{RBLFailLog};
 $Config{OutgoingBufSize}=102400 unless $Config{OutgoingBufSize};
 $Config{MaxBytes}=10000 unless $Config{MaxBytes};
 $Config{AvError}='500 Mail appears infected with \'$infection\' -- disinfect and resend.' unless $Config{AvError};
 $Config{LocalPolicySPF}='v=spf1 a/24 mx/24 ptr ~all' unless $Config{LocalPolicySPF};
  $Config{RBLServiceProvider}='bl.spamcop.net|cbl.abuseat.org|sbl-xbl.spamhaus.org|dnsbl.njabl.org|list.dsbl.org|dnsbl.sorbs.net|opm.blitzed.org|dynablock.njabl.org' unless $Config{RBLServiceProvider};

 $Config{RBLmaxtime}=10 unless $Config{RBLmaxtime};
 $Config{RBLsocktime}=1 unless $Config{RBLsocktime};

 # for upgrading from earlier than 1.1.1
 if ($Config{TestMode}) {
    $Config{baysTestMode}=1;
    $Config{blTestMode}=1;
    $Config{hlTestMode}=1;   
    $Config{sbTestMode}=1;
    $Config{spfTestMode}=1;
    $Config{rblTestMode}=1;
 }
 if (exists($Config{TestMode})) {
		delete($Config{TestMode});
 }
 $Config{BadAttachL1}=$Config{ExtensionsToBlock} unless $Config{BadAttachL1};
 $Config{freqSpam}=1 unless $Config{freqSpam};
 $Config{freqNonSpam}=1 unless $Config{freqNonSpam};
 $Config{LogRollDays}=14 unless $Config{LogRollDays};
    # -- this sets the variable name with the same name as the config key to the new value
    # -- for example $Config{myName}="ASSP-nospam" -> $myName="ASSP-nospam";
 for (keys %Config) {${$_}=$Config{$_};}

 $EmailFrom="<>" unless $EmailFrom;
  # turn settings into regular expressions
 @PossibleOptionFiles=();
 for (@Config) {
  if($_->[6] eq 'ConfigMakeRe') {
   ${$_->[0]}=optionList(${$_->[0]},$_->[0]);
   push(@PossibleOptionFiles,$_->[0]);
  } elsif($_->[6] eq 'ConfigCompileRe') {
   ConfigCompileRe($_->[0],'',${$_->[0]},'Initializing');
  }
 }
 updateBadAttachL1(BadAttachL1,'',$Config{BadAttachL1},'Initializing');
 updateRBL(updateRBL,'',$Config{ValidateRBL},'Initializing');
 updateLog2(updateLog2,'',$Config{freqSpam},'Initializing');
 SaveConfig();
}

sub ConfigMakeRe { my ($name, $old, $new)=@_;
 $$name=optionList($new,$name);
 mlog(0,"Admin update: $name changed from '$old' to '$new'");
 '';
}

sub ConfigCompileRe { my ($name, $old, $new, $init)=@_;
 $$name=$new;
 my $n2=$name."RE";
 if($new) {
  SetRE($n2,$new,"is",$name);
 }
 mlog(0,"Admin update: $name changed from '$old' to '$new'") unless $init;
 '';
}

sub optionList {
 # this converts a | separated list into a RE and handles the options in a file
 my $d=shift;
 my $configname=shift;
 if($d=~/^ *file: *(.+)/i) {
  # the optionlist is actually saved in a file.
  my $fil=$1;
  local $/;
  my @s=stat($fil);
  my $mtime=$s[9];
  $FileUpdate{$fil}=$mtime;
  if(open(OL,"<$fil")) {
   $d=<OL>; # read the file
   $d=~s/ #.*//g; # clean off comments
   $d=~s/\s*\n\s*/|/g; # replace newlines (and the whitespace that surrounds them) with a |
   mlog(0,"Option list file '$fil' reloaded ($configname).");
   close OL;
  } else {
   mlog(0,"Failed to open option list file '$fil' ($configname): $!");
  }
 }
 $d=~s/\|\|+/\|/g;
 $d=~s/^\|//;
 $d=~s/\|$//;
 $d=~s/([\.\[\]\-\(\)\*\+\\])/\\$1/g;
 $MakeRE{$configname}->($d);
 $d;
}

sub fileUpdated {
 my $fil=shift;
 return 1 unless $FileUpdate{$fil};
 my @s=stat($fil);
 my $mtime=$s[9];
 $FileUpdate{$fil}!=$mtime;
}

sub ConfigChangeMailPort {my ($name, $old, $new)=@_;
 if($> == 0 || $new >= 1024) {
  # change the listenport
  $listenPort=$new;
  if($lsn) {
   $readable->remove($lsn);
   $lsn->close();
  }
  $lsn = newListen($listenPort,\&NewSMTPConnection);
  mlog(0,"Listening on new mail port $listenPort (changed from $old) per admin request");
  return '';
 } else {
  # don't have permissions to change
  mlog(0,"Request to listen on new mail port $listenPort (changed from $old) -- restart required; euid=$>");
  return "<br />Restart required; euid=$>";
 }
}

sub ConfigChangeMailPort2 {my ($name, $old, $new)=@_;
 if($> == 0 || $new >= 1024) {
  # change the listenport2
  $listenPort2=$new;
  if($lsn2) {
   $readable->remove($lsn2);
   $lsn2->close();
  }
  $lsn2 = newListen($listenPort2,\&NewSMTPConnection);
  mlog(0,"Listening on new secondary mail port $listenPort2 (changed from $old) per admin request");
  return '';
 } else {
  # don't have permissions to change
  mlog(0,"Request to listen on new secondary mail port $listenPort2 (changed from $old) -- restart required; euid=$>");
  return "<br />Restart required; euid=$>";
 }
}

sub ConfigChangeAdminPort {my ($name, $old, $new)=@_;
 if($> == 0 || $new >= 1024) {
  # change the listenport
  $webAdminPort=$new;
  $readable->remove($WebSocket);
  $WebSocket->close();
  $WebSocket = newListen($webAdminPort,\&NewWebConnection);
  if($WebSocket) {
   mlog(0,"Listening on new admin port $new (changed from $old) per admin request");
  } else {
   # couldn't open the port -- switch back
   $webAdminPort=$old;
   $WebSocket = newListen($webAdminPort,\&NewWebConnection);
   mlog(0,"Couldn't open new port -- still listening on $old");
   $Config{$name}=$old;
   return "<span class=\"negative\">Couldn't open new port $new</span>";
  }
  return '';
 } else {
  # don't have permissions to change
  mlog(0,"Request to listen on new admin port $new (changed from $old) -- restart required; euid=$>");
  return "<br />Restart required; euid=$>";
 }
}

sub ConfigChangeRelayPort {my ($name, $old, $new)=@_;
 unless ($relayHost && $new) {
  if($Relay) {
   $readable->remove($Relay);
   $Relay->close();
   mlog(0,"Relay port disabled");
   return '<br />relay port disabled';
  } else {
   return "<br />relayHost ($relayHost) and relayPort ($new) must be defined to enable relaying";
  }
 }
 if($> == 0 || $new >= 1024) {
  # change the listenport
  $relayPort=$new;
  if($Relay) {
   $readable->remove($Relay);
   $Relay->close();
  }
  $Relay=newListen($relayPort,\&NewSMTPConnection);
  mlog(0,"Listening for relay connections at $relayPort -- changed per admin request");
  return '';
 } else {
  # don't have permissions to change
  mlog(0,"Request to listen on new relay port $new (changed from $old) -- restart required; euid=$>");
  return "<br />Restart required; euid=$>";
 }
}

sub ConfigChangeLogfile {my ($name, $old, $new)=@_;
 close LOG if $logfile;
 $logfile=$new;
 if($logfile && open(LOG,">>$base/$logfile")) {my $oldfh = select(LOG); $| = 1; select($oldfh);}
 mlog(0,"Log file changed to '$new' from '$old' per admin request");
 '';
}
sub ConfigDEBUG {my ($name, $old, $new)=@_;
 close DEBUG if $DEBUG;
 $DEBUG=$new;
 if($DEBUG) {
  open(DEBUG, ">$base/".time.".dbg"); binmode(DEBUG); my $oldfh = select(DEBUG); $| = 1; select($oldfh);
  eval(q[sub d {
   $time=gmtime(); $time=~s/... (...) (..) (........) ..(..)/$2 $1 $4 $3/;
   print DEBUG "$time <$_[0]>";
  }
  ]);
 } else {
  eval(q[sub d{}]);
 }
 mlog(0,"Debug file changed to '$new' from '$old' per admin request");
 '';
}

# Bad Attachment Settings, Checks and Update.
sub updateBadAttachL1 {my ($name, $old, $new, $init)=@_;
  mlog(0,"Badattach Level 1 updated from '$old' to '$new'.") unless $init;
  SetRE(badattachL1RE,qq[content-\\w+: .*\\s+.*name\\s*=\\s*".*\\.($new)"|content-\\w+: .*\\s+.*name\\s*=\\s*.*\\.($new)\\s],'i',"Bad Attachment L1");
 updateBadAttachL2(BadAttachL2,'',$Config{BadAttachL2},$new);
}
sub updateBadAttachL2 {my ($name, $old, $new, $init)=@_;
  mlog(0,"Badattach Level 2 updated from '$old' to '$new'.") unless $init;
  $new.='|'.$init;
  SetRE(badattachL2RE,qq[content-\\w+: .*\\s+.*name\\s*=\\s*".*\\.($new)"|content-\\w+: .*\\s+.*name\\s*=\\s*.*\\.($new)\\s],'i',"Bad Attachment L2");
 updateBadAttachL3(BadAttachL3,'',$Config{BadAttachL3},$new);
}
sub updateBadAttachL3 {my ($name, $old, $new, $init)=@_;
 mlog(0,"Badattach Level 3 updated from '$old' to '$new'.") unless $init;
 $new.='|'.$init;
 SetRE(badattachL3RE,qq[content-\\w+: .*\\s+.*name\\s*=\\s*".*\\.($new)"|content-\\w+: .*\\s+.*name\\s*=\\s*.*\\.($new)\\s],'i',"Bad Attachment L2");
 $badattachRE[1]=$badattachL1RE;
 $badattachRE[2]=$badattachL2RE;
 $badattachRE[3]=$badattachL3RE;
 '';
}

# RBL Settings Checks, and Update.
sub updateRBL {my ($name, $old, $new, $init)=@_;
  mlog(0,"RBL-Enable updated from '$old' to '$new'.") unless $init;
  $ValidateRBL=$Config{ValidateRBL}=$new;
  if (!$CanUseRBL) {
   mlog(0,"RBL-Enable updated from '1' to ''.") if $Config{ValidateRBL};
   $ValidateRBL=$Config{ValidateRBL}=undef;
   return '<span class="negative">*** Net::DNS must be installed before enabling RBL.</span>';
  } else {
   updateRBLMH(updateRBLMH,'',$Config{RBLmaxhits},'Cascading');
  }
}
sub updateRBLMH {my ($name, $old, $new, $init)=@_;
 mlog(0,"RBL Maximum Hits updated from '$old' to '$new'.") unless $init;
 $RBLmaxhits=$new;
 if ($new<=0) {
  mlog(0,"RBL-Enable updated from '1' to ''.") if $Config{ValidateRBL};
  $ValidateRBL=$Config{ValidateRBL}=undef;
  return '<span class="negative">*** RBLmaxhits must be defined and positive before enabling RBL.</span>';
 } else {
  updateRBLMR(updateRBLMR,'',$Config{RBLmaxreplies},'Cascading');
 }
}
sub updateRBLMR {my ($name, $old, $new, $init)=@_;
 mlog(0,"RBL Maximum Replies updated from '$old' to '$new'.") unless $init;
 $RBLmaxreplies=$new;
 if ($new<$RBLmaxhits) {
  mlog(0,"RBL-Enable updated from '1' to ''.") if $Config{ValidateRBL};
  $ValidateRBL=$Config{ValidateRBL}=undef;
  return '<span class="negative">*** RBLmaxreplies must be more than or equal to RBLmaxhits before enabling RBL.</span>';
 } else {
  updateRBLSP(updateRBLSP,'',$Config{RBLServiceProvider},'Cascading');
 }
}
sub updateRBLSP {my ($name, $old, $new, $init)=@_;
 mlog(0,"RBL Service Providers updated from '$old' to '$new'.") unless $init;
 $RBLServiceProvider=$new;
 my $domains = ($new =~ s/\|/$1/g)+1;
 if ($domains<$RBLmaxreplies) {
  mlog(0,"RBL-Enable updated from '1' to ''.") if $Config{ValidateRBL};
  $ValidateRBL=$Config{ValidateRBL}=undef;
  return '<span class="negative">*** RBLServiceProvider must contain more than or equal to RBLmaxreplies domains before enabling RBL.</span>';
 } elsif($CanUseRBL) {
  my $res = Net::DNS::Resolver->new;
  @nameservers = $res->nameservers;
  @rbllist = split(/\|/,$RBLServiceProvider);
  if ($init && $ValidateRBL) {
   return ' & RBL activated';
  } else {
   return '';
  }
 }
}

# Database File Logging Frequency Setup.
sub updateLog2 {my ($name, $old, $new, $init)=@_;
 mlog(0,"Non Spam Logging Frequency updated from '$old' to '$new'.") unless $init;
 $logFreq[2]=$new;
 updateLog3(updateLog3,'',$Config{freqNonSpam},$new);
}
sub updateLog3 {my ($name, $old, $new, $init)=@_;
 mlog(0,"Spam Logging Frequency updated from '$old' to '$new'.") unless $init;
 $logFreq[3]=$new;
 return '';
}

sub reloadConfigFile {
 # called on SIG HUP
 my %newConfig;
 mlog(0,"Sig HUP -- reloading config");
 open(F,"<$base/assp.cfg"); local $/; (%newConfig)=split(/:=|\n/,<F>); close F;
 for $c (@Config) {
  my ($name,$nicename,$size,$func,$default,$valid,$onchange,$description)=@$c;
  if($Config{$name} ne $newConfig{$name}) {
   if($newConfig{$name}=~/$valid/i) {
    my $new=$1; my $info;
    if($onchange) {
     $info=$onchange->($name,$Config{$name},$new);
    } else {
     mlog(0,"Admin update: $name changed from '$Config{$name}' to '$new'");
     ${$name}=$new;
     # -- this sets the variable name with the same name as the config key to the new value
     # -- for example $Config{myName}="ASSP-nospam" -> $myName="ASSP-nospam";
    }
    $Config{$name}=$new;
   } else {
    mlog(0, "Error: invalid '$newConfig{$name}' -- not changed");
   }
  }
 }
  for $f (@PossibleOptionFiles) {
   ${$f}=optionList($Config{$f},$f) if $Config{$f}=~/^ *file: *(.+)/i && fileUpdated($1);
  }
 # reopen log file, just for fun.
  close LOG;
  if(open(LOG,">>$base/$logfile")) {my $oldfh = select(LOG); $| = 1; select($oldfh);}
  print LOG "Logfile reopened on HUP\n";
}

#####################################################################################
#                orderedtie
{
package orderedtie;
# This is a tied value that caches lookups from a sorted file; \n separates records,
# \002 separates the key from the value. After main::OrderedTieHashSize lookups the cache is
# cleared. This give us most of the speed of the hash without the huge memory overhead of storing
# the entire hash and should be totally portable. Picking the best value for n requires some
# tuning. A \n is required to start the file.

# if you're updating entries it behoves you to call flush every so often to make sure that your
# changes are saved. This also frees the memory used to remember updated values.

# for my purposes a value of undef and a nonexistant key are the same

# Obviously if your keys or values contain \n or \002 it will totally goof things up.


sub TIEHASH {
 my ($c,$fn)=@_;
 my $self={
  fn => $fn,
  age => mtime($fn),
  cnt => 0,
  cache => {},
  updated => {},
  ptr => 1,
 };
 bless $self, $c;
 return $self;
}
sub DESTROY { $_[0]->flush(); }

sub mtime { my @s=stat($_[0]); $s[9]; }

sub flush {
 my $this=shift;
 return unless %{$this->{updated}};
 my $f=$this->{fn};
 open(O,">$f.tmp") || return undef;
 binmode(O);
 open(I,"<$f") || print O "\n";
 binmode(I);
 local $/="\n";
 my @l=(sort keys %{$this->{updated}});
 my ($k,$d,$r,$v);
 while($r=<I>) {
  ($k,$d)=split("\002",$r);
  while(@l && $l[0] lt $k) {
   $v=$this->{updated}{$l[0]};
   print O "$l[0]\002$v\n" if $v;
   shift(@l);
  }
  if($l[0] eq $k) {
   $v=$this->{updated}{$l[0]};
   print O "$l[0]\002$v\n" if $v;
   shift(@l);
  } else {
   print O $r;
  }
 }
 while(@l) {
  $v=$this->{updated}{$l[0]};
  print O "$l[0]\002$v\n" if $v;
  shift(@l);
 }
 close I; close O; unlink($f); rename("$f.tmp", $f);
 $this->{updated}={};
}

sub STORE {
 my ($this, $key, $value)=@_;
 $this->{cache}{$key}=$this->{updated}{$key}=$value;
}

sub FETCH { my ($this, $key)=@_;
 return $this->{cache}{$key} if exists $this->{cache}{$key};
 $this->resetCache() if($this->{cnt}++ >$main::OrderedTieHashSize || ($this->{cnt} & 0x1f) == 0 && mtime($this->{fn}) != $this->{age});

 return $this->{cache}{$key}=binsearch($this->{fn},$key);
}

sub resetCache {
 my $this=shift;
 $this->{cnt}=0;
 $this->{age}=mtime($this->{fn});
 $this->{cache}={%{$this->{updated}}};
 main::mlog(0,"cache reset ($this->{fn})");
}

sub binsearch {
 my ($f,$k)=@_;
 open(F,"<$f") || return undef;
 binmode(F);
 my $count=0;
 my $siz=my $h=-s $f;
 $siz-=1024;
 my $l=0;
 my $k0=$k;
 $k=~s/([\[\]\(\)\*\^\!\|\+\.\\\/\?\`\$\@\{\}])/\\$1/g; # make sure there's no re chars unqutoed in the key
 while(1) {
  my $m=(($l+$h)>>1)-1024;
  $m=0 if $m < 0;
  seek(F,$m,0);
  my $d; my $read= read(F,$d,2048);
  if( $d=~/\n$k\002([^\n]*)\n/) {
   close F;
   return $1;
  }
  my ($pre,$first,$last,$post)=$d=~/^(.*?)\n(.*?)\002.*\n(.*?)\002.*?\n(.*?)$/s;
  last unless defined $first;
  if($k0 gt $first && $k0 lt $last) {
   last;
  }
  if($k0 lt $first) {
   last if $m ==0;
   $h=$m-1024+length($pre);
   $h=0 if $h < 0;
  }
  if($k0 gt $last) {
   last if $m >= $siz;
   $l=$m+$read-length($post);
  }
  if($count++ > 100) {
   main::mlog(0,"Error: $this->{fn} must be repaired ($k0)");
   last;
  }
 }
 close F;
 return undef;
}

sub FIRSTKEY { $this=shift;
 $this->flush();
 $this->{ptr}=1;
 $this->NEXTKEY();
}
sub NEXTKEY { my ($this, $lastkey)=@_;
 local $/="\n";
 open(F,"<$this->{fn}") || return undef;
 binmode(F);
 seek(F,$this->{ptr},0);
 my $r=<F>;
 return undef unless $r;
 $this->{ptr}=tell F;
 close F;
 my ($k,$v)=$r=~/(.*?)\002(.*?)\n/s;
 if(!exists($this->{cache}{$k}) && $this->{cnt}++ >$main::OrderedTieHashSize) {
  $this->{cnt}=0;
  $this->{cache}={%{$this->{updated}}};
 }
 $this->{cache}{$k}=$v;
 $k;
}

sub EXISTS { my ($this, $key)=@_;
 return FETCH($this, $key);
}

sub DELETE {my ($this, $key)=@_;
 $this->{cache}{$key}=$this->{updated}{$key}=undef;
}

sub CLEAR {my ($this)=@_;
 open(F,">$this->{fn}"); binmode(F); print F "\n"; close F;
 $this->{cache}={};
 $this->{updated}={};
 $this->{cnt}=0;
}
}

{
#################################################################
# this package implements a pure perl virus scanner
# it uses the clam anti-virus databases (see www.clamav.net)
# Download your databases this way -- maybe once a day:
#  wget --timestamping http://database.clamav.net/database/viruses.db
#  wget --timestamping http://database.clamav.net/database/viruses.db2
#
# copyright (C) 2004, John Hanna under the terms of the GPL

package Av;

# load the databases -- return the number of signatures present
# optional parameters:
#   path => "/path/to/your/clamav/db/files", # default .
#   databases => "viruses.db,viruses.db2" # as many or as few as you want
sub init { my($proto,$args)=@_;
 $path= $args->{path} || ".";
 $databases = $args->{databases} || "viruses.db,viruses.db2";
 loadAll();
}

# public function to create a new scan buffer -- see addchar below
sub new { bless({offset=>0, buf=>'', prereq=>{}}, ref($_[0]) || $_[0]); }

# public function to reset the scan buffer
sub clear {my $self=shift; $self->{offset}=0; $self->{buf}=''; $self->{prereq}={};}

# return the number of signatures in the virus database
sub count {$count}

# called internally, but can be called to manually reload the virus signature database
sub loadAll {
 $count=0;
 $longest=0;
 undef %prereqs;
 $prereqcount='a';
 undef @db;
 load("$path/$_") for (split(/,/,$databases));
 $count;
}

# called internally to add a signature file to the database
sub load { my ($f)=@_;
 open(F,"<$f");
 $fileTimes{$f}=[stat(F)]->[9];
 my ($nam,$sig,$lsig);
 while(<F>) {
  ($nam,$sig)=/(.*)=(.*)/;
  next unless $sig;
  $sig=lc $sig;
  $nam=~s/ \(clam\)//i;
  $nam=~s/'/\\'/g;
  my @parts=split(/\*/,$sig);
  my $lp;
  for my $part (0 .. $#parts) {
   my $p=$parts[$part];
   $longest = length($p)/2 if length($p)/2 > $longest;
   my $prereqn=$prereqs{$p} || ($prereqs{$p}=++$prereqcount);# if($part < $#parts);
   my $prereq=$part? $prereqs{$lp} :'';
   $lp=$p;
   my $return=$part == $#parts? $nam:'';
   my $setprereq=$part == $#parts? '' : $prereqn;
   my ($tail)=$p=~/([0-9a-f]{2,16})$/;
   my $ltail = length($tail);
   my $hsize=$ltail >=16? 8: $ltail >= 8? 4: $ltail >= 4? 2: 1;
   my $loc=substr($tail,-2*$hsize);

   my($pretest,$rest)=$p=~/([0-9a-f]{2,8})([0-9a-f\?]*)$/;
   $pretest=pack("H*",$pretest);
   my $lpretest=length($pretest);
   $rest=length($rest)/2+$lpretest;
   my @pretest;
   if($rest>$hsize && length($p) > 32) {
    @pretest=(-$rest,$lpretest,$pretest);
   } else {@pretest=(0,0,'');}
   $test=$p;
   $test=~s/(..)/\\x$1/g;
   my ($re,$match,$matchlen)=('','',0);
   if($test=~s/\\x\?\?/./g) {
    $re=qr/$test$/s;
   } else {
    my $len=length($p)/2;
    $matchlen=$len; $match=pack("H*",$p);
    #$match="" if $len == $hsize;
   }

   push(@{$db[$hsize]->{pack("H*",$loc)}}, [$prereq,$return,$setprereq,@pretest,$re,$match,$matchlen]);
  }
  $count++;
 }
 close F;
}

# public function to check if the database needs to be reloaded
# use this this way: Av->loadAll if Av->checkReload;
sub checkReload {
 for my $f (split(/,/,$databases)) {
  $f="$path/$f";
  # we want to reload if a file is newer (has later date) than when read,
  # and is at least 120 seconds old -- in lieu of locking
  my $modtime=[stat($f)]->[9];
  if($modtime > $fileTimes{$f} && time - $modtime > 120) {
   return 1;
  }
 }
 0;
}

# public function to scan a file
# returns undef if no virus found
# returns array-ref with the offset into the buffer and the name of the virus if found
# you can pass in an offset to start scanning to continue scanning the file
# you may need to $av->clear to get the desired effects
sub scanfile { my ($self, $file, $n)=@_;
 open(F,"<$file") || die("Couldn't open $file: $!"); binmode F;
 my $c;
 $self->{offset}=$n; $self->{buf}=''; $self->{prereq}={};
 seek(F,$n,0) if $n>0;
 seek(F,$n,2) if $n < 0;
 my $r;
 while(defined ($c=getc(F))) {
  return $r if $r=$self->addchar($c);
 }
 close F;
 undef;
}

# public function to do character-at-a-time scanning.
# $av=Av->new();
# for my $c (split(//,$buf)) {
#  print "'$r->[1]' virus found at offset $r->[0]\n" if $r=$av->addchar($c);
# }
sub addchar {
 my $self=shift;
 $self->{buf} .= $_[0];
 $self->{offset}++;
 if(length($self->{buf}) > $longest * 2) {
  $self->{buf}=substr($self->{buf}, -$longest);
 }
 for $hsiz (8,4,2,1) {
  my $l=$db[$hsiz]->{substr($self->{buf},-$hsiz)};
  for (@{$l}){
   if( ! $_->[0] || $self->{prereq}->{$_->[0]} ) {
    # pre test
    if(! $_->[4] || substr($self->{buf},$_->[3],$_->[4]) eq $_->[5]) {
     # real test
     if($_->[8] && substr($self->{buf},-$_->[8]) eq $_->[7] || ! $_->[8] && $self->{buf}=~$_->[6]) {
      $_->[2] && ($self->{prereq}->{$_->[2]}=1) || return [$self->{offset},$_->[1]];
     }
    }
   }
  }
 }
 return undef;
}

}

#################################################################
# this package implements realtime blacklisting
# it is based on Net::RBLClient by Asher Blum <asher@wildspark.com>
# CREDITS Martin H. Sluka <martin@sluka.de>
# Copyright (C) 2002 Asher Blum.  All rights reserved.
# This code is free software; you can redistribute it and/or modify it under
# the same terms as Perl itself.
# Modified for integration with ASSP by John Calvi.


package RBL;

use IO::Socket;

sub new {

    # This avoids compile time errors if Net::DNS is not installed.
    # The error will be returned on the lookup function call.
    if ($main::CanUseDNS) {
       require Net::DNS::Packet;
       $CanUseDNS=1;
    }  

    my($class, %args) = @_;
    my $self = {
        lists       => [ lists() ],
        query_txt   => 1,
        max_time    => 10,
        timeout     => 1,
        max_hits    => 3,
        max_replies => 6,
        udp_maxlen  => 4000,
        server      => '127.0.0.1',
    };
    bless $self, $class;
    foreach my $key(keys %args) {
        defined($self->{ $key })
            or return "Invalid key: $key";
        $self->{ $key } = $args{ $key }; 
    }

    $self;
}

sub lookup {
    return "Net::DNS package required" unless $CanUseDNS;
    my($self, $target_ip) = @_;
    my $ip_pat = qr(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3});
    $target_ip =~ /$ip_pat/
        or return "Invalid ip: '$target_ip' - must be dotted quad";
    my $start_time = time;
    my $qip = join '.', reverse(split /\./, $target_ip);
    my $deadline = time + $self->{ max_time };

    my $sock = IO::Socket::INET->new(
        Proto     => 'udp',
        PeerPort  => 53,
        PeerAddr  => $self->{ server },
    ) or return "Failed to create UDP client";

    if ( $self->{ query_txt } ) {
      foreach my $list(@{ $self->{ lists } }) {
        my($msg_a, $msg_t) = mk_packet($qip, $list);
        foreach ($msg_a, $msg_t) { 
          $sock->send($_) or return "send: $!";
        }
      }
    } else {
        foreach my $list(@{ $self->{ lists } }) {
          my $msg = mk_packet($qip, $list);
          $sock->send($msg) || return "send: $!";
        }
    }

    $self->{ results } = {};
    $self->{ txt } = {};

    my $needed = 0;
    if ($self->{ max_replies} > @{ $self->{ lists } }) {
      $needed = @{ $self->{ lists } }; 
    } else {
      $needed = $self->{ max_replies };
    }
    $needed <<= 1 if $self->{ query_txt }; # how many packets needed back
    my $hits = my $replies = 0;

    my $select = new IO::Select;
    $select->add($sock);

    # Keep receiving packets until one of the exit conditions is met: 
    main::mlog(0,"Commencing RBL checks on $target_ip") if $main::RBLLog;
    while($needed && time < $deadline) {
      my $msg = '';
      my $dur = time - $start_time;
      if( my @ready = $select->can_read($self->{ timeout }) ){
        $sock->recv($msg, $self->{ udp_maxlen })  || return "recv: $!";
      } else {
        # waiting for other replies.
      }
      if($msg) {
        my ($domain, $res, $type) = decode_packet($msg);
        if ( defined $type && $type eq 'TXT' ) {
          $self->{ txt }{ $domain } = $res
        }
        elsif ($res) {
          $replies ++;
          $hits ++ if $res;
          $self->{ results }{ $domain } = $res;
          return 1 if    $hits >= $self->{ max_hits } ||
                         $replies >= $self->{ max_replies };
        }
        $needed --;
      }
    }
    main::mlog(0,"Completed RBL checks on $target_ip") if $main::RBLLog;
    close ($sock);
    1;
}

sub listed_by {
    my $self = shift;
    sort keys %{ $self->{ results } };
}

sub listed_hash {
    my $self = shift;
    %{ $self->{ results } };
}

sub txt_hash {
    my $self = shift;
    warn <<_ unless $self->{ query_txt };
Without query_txt turned on, you won't get any results from ->txt_hash().
_
    if (wantarray) { %{ $self->{ txt } } }
    else { $self->{ txt } }
}

# End methods - begin internal functions

sub mk_packet {
    # pass me a REVERSED dotted quad ip (qip) and a blocklist domain
    my($qip, $list) = @_;
    my($packet, $error) = new Net::DNS::Packet(my $fqdn = "$qip.$list", 'A');
    return "Cannot build DNS query for $fqdn, type A: $error" unless $packet;
    return $packet->data unless wantarray;
    my($txt_packet, $error) = new Net::DNS::Packet($fqdn, 'TXT', 'IN');
    return "Cannot build DNS query for $fqdn, type TXT: $error" unless $txt_packet;
    $packet->data, $txt_packet->data;
}

sub decode_packet {
    # takes a raw DNS response packet
    # returns domain, response
    my $data = shift;
    my $packet = Net::DNS::Packet->new(\$data);
    my @answer = $packet->answer;

    {
        my($res, $domain, $type);
        foreach my $answer (@answer) {
            {
                my $name = lc $answer->name;
                warn $answer->answerfrom .
                  " returned answers to different domains ($domain and $answer)"
                  if defined $domain && $name ne $domain;
                $domain = $answer->name;
            }
            $domain =~ s/^\d+\.\d+\.\d+\.\d+\.//;
            $type = $answer->type;
            $res = $type eq 'A'     ? inet_ntoa($answer->rdata)  :
                   $type eq 'CNAME' ?   cleanup($answer->rdata)  :
                   $type eq 'TXT'   ? (defined $res && "$res; ")
                                      . $answer->txtdata         :
                   '?';
            last unless $type eq 'TXT';
        }
        return $domain, $res, $type if defined $res;
    }
    
    # OK, there were no answers -
    # need to determine which domain
    # sent the packet.

    my @question = $packet->question;
    foreach my $question(@question) {
        my $domain = $question->qname;
        $domain =~ s/^\d+\.\d+\.\d+\.\d+\.//;
        return($domain, undef);
    }
}

sub cleanup {
    # remove control chars and stuff
    $_[ 0 ] =~ tr/a-zA-Z0-9./ /cs;;
    $_[ 0 ];
}


sub lists {
    qw(
       bl.spamcop.net
       cbl.abuseat.org
       sbl-xbl.spamhaus.org
       dnsbl.njabl.org
       list.dsbl.org
       dnsbl.sorbs.net
       opm.blitzed.org
       dynablock.njabl.org
    );
}
1;
